# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim_test_helpers import *

testseed = 1000

precision = 10
# decimal point at which agreement is required for all double precision tests

precisionD = precision
precisionF = 5  # precision=10 does not make sense at single precision
precisionS = 1  # "precision" also a silly concept for ints, but allows all 4 tests to run in one go
precisionI = 1

@timer
def test_deviate_noise():
    """Test basic functionality of the DeviateNoise class
    """
    u = galsim.UniformDeviate(testseed)
    uResult = np.empty((10,10))
    u.generate(uResult)

    noise = galsim.DeviateNoise(galsim.UniformDeviate(testseed))

    # Test filling an image with random values
    testimage = galsim.ImageD(10,10)
    testimage.addNoise(noise)
    np.testing.assert_array_almost_equal(
            testimage.array, uResult, precision,
            err_msg='Wrong uniform random number sequence generated when applied to image.')

    # Test filling a single-precision image
    noise.rng.seed(testseed)
    testimage = galsim.ImageF(10,10)
    testimage.addNoise(noise)
    np.testing.assert_array_almost_equal(
            testimage.array, uResult, precisionF,
            err_msg='Wrong uniform random number sequence generated when applied to ImageF.')

    # Test filling an image with Fortran ordering
    noise.rng.seed(testseed)
    testimage = galsim.ImageD(np.zeros((10,10)).T)
    testimage.addNoise(noise)
    np.testing.assert_array_almost_equal(
            testimage.array, uResult, precision,
            err_msg="Wrong uniform randoms generated for Fortran-ordered Image")

    # Check picklability
    check_pickle(noise, drawNoise)
    check_pickle(noise)

    # Check copy, eq and ne
    noise2 = galsim.DeviateNoise(noise.rng.duplicate())  # Separate but equivalent rng chain.
    noise3 = noise.copy()  # Always has exactly the same rng as noise.
    noise4 = noise.copy(rng=galsim.BaseDeviate(11))  # Always has a different rng than noise
    assert noise == noise2
    assert noise == noise3
    assert noise != noise4
    assert noise.rng() == noise2.rng()
    assert noise == noise2  # Still equal because both chains incremented one place.
    assert noise == noise3  # Still equal because noise 3's rng is always equal to noise's rng.
    noise.rng()
    assert noise2 != noise3 # This is no longer equal, since only noise.rng is incremented.
    assert noise == noise3

    assert_raises(TypeError, galsim.DeviateNoise, 53)
    assert_raises(NotImplementedError, galsim.BaseNoise().getVariance)
    assert_raises(NotImplementedError, galsim.BaseNoise().withVariance, 23)
    assert_raises(NotImplementedError, galsim.BaseNoise().withScaledVariance, 23)
    assert_raises(TypeError, noise.applyTo, 23)
    assert_raises(NotImplementedError, galsim.BaseNoise().applyTo, testimage)
    assert_raises(galsim.GalSimError, noise.getVariance)
    assert_raises(galsim.GalSimError, noise.withVariance, 23)
    assert_raises(galsim.GalSimError, noise.withScaledVariance, 23)


@timer
def test_gaussian_noise():
    """Test Gaussian random number generator
    """
    gSigma = 17.23
    g = galsim.GaussianDeviate(testseed, sigma=gSigma)
    gResult = np.empty((10,10))
    g.generate(gResult)
    noise = galsim.DeviateNoise(g)

    # Test filling an image
    testimage = galsim.ImageD(10,10)
    noise.rng.seed(testseed)
    testimage.addNoise(noise)
    np.testing.assert_array_almost_equal(
            testimage.array, gResult, precision,
            err_msg='Wrong Gaussian random number sequence generated when applied to image.')

    # Test filling a single-precision image
    noise.rng.seed(testseed)
    testimage = galsim.ImageF(10,10)
    testimage.addNoise(noise)
    np.testing.assert_array_almost_equal(
            testimage.array, gResult, precisionF,
            err_msg='Wrong Gaussian random number sequence generated when applied to ImageF.')

    # GaussianNoise is equivalent, but no mean allowed.
    gn = galsim.GaussianNoise(galsim.BaseDeviate(testseed), sigma=gSigma)
    testimage = galsim.ImageD(10,10)
    testimage.addNoise(gn)
    np.testing.assert_array_almost_equal(
            testimage.array, gResult, precision,
            err_msg="GaussianNoise applied to Images does not reproduce expected sequence")

    # Test filling an image with Fortran ordering
    gn.rng.seed(testseed)
    testimage = galsim.ImageD(np.zeros((10,10)).T)
    testimage.addNoise(gn)
    np.testing.assert_array_almost_equal(
            testimage.array, gResult, precision,
            err_msg="Wrong Gaussian noise generated for Fortran-ordered Image")

    # Check GaussianNoise variance:
    np.testing.assert_almost_equal(
            gn.getVariance(), gSigma**2, precision,
            err_msg="GaussianNoise getVariance returns wrong variance")
    np.testing.assert_almost_equal(
            gn.sigma, gSigma, precision,
            err_msg="GaussianNoise sigma returns wrong value")

    # Check that the noise model really does produce this variance.
    big_im = galsim.Image(2048,2048,dtype=float)
    gn.rng.seed(testseed)
    big_im.addNoise(gn)
    var = np.var(big_im.array)
    print('variance = ',var)
    print('getVar = ',gn.getVariance())
    np.testing.assert_almost_equal(
            var, gn.getVariance(), 1,
            err_msg='Realized variance for GaussianNoise did not match getVariance()')

    # Check that GaussianNoise adds to the image, not overwrites the image.
    gal = galsim.Exponential(half_light_radius=2.3, flux=1.e4)
    gal.drawImage(image=big_im)
    gn.rng.seed(testseed)
    big_im.addNoise(gn)
    gal.withFlux(-1.e4).drawImage(image=big_im, add_to_image=True)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(
            var, gn.getVariance(), 1,
            err_msg='GaussianNoise wrong when already an object drawn on the image')

    # Check that DeviateNoise adds to the image, not overwrites the image.
    gal.drawImage(image=big_im)
    gn.rng.seed(testseed)
    big_im.addNoise(gn)
    gal.withFlux(-1.e4).drawImage(image=big_im, add_to_image=True)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(
            var, gn.getVariance(), 1,
            err_msg='DeviateNoise wrong when already an object drawn on the image')

    # Check withVariance
    gn = gn.withVariance(9.)
    np.testing.assert_almost_equal(
            gn.getVariance(), 9, precision,
            err_msg="GaussianNoise withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            gn.sigma, 3., precision,
            err_msg="GaussianNoise withVariance results in wrong sigma")

    # Check withScaledVariance
    gn = gn.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            gn.getVariance(), 36., precision,
            err_msg="GaussianNoise withScaledVariance results in wrong variance")
    np.testing.assert_almost_equal(
            gn.sigma, 6., precision,
            err_msg="GaussianNoise withScaledVariance results in wrong sigma")

    # Check arithmetic
    gn = gn.withVariance(0.5)
    gn2 = gn * 3
    np.testing.assert_almost_equal(
            gn2.getVariance(), 1.5, precision,
            err_msg="GaussianNoise gn*3 results in wrong variance")
    np.testing.assert_almost_equal(
            gn.getVariance(), 0.5, precision,
            err_msg="GaussianNoise gn*3 results in wrong variance for original gn")
    gn2 = 5 * gn
    np.testing.assert_almost_equal(
            gn2.getVariance(), 2.5, precision,
            err_msg="GaussianNoise 5*gn results in wrong variance")
    np.testing.assert_almost_equal(
            gn.getVariance(), 0.5, precision,
            err_msg="GaussianNoise 5*gn results in wrong variance for original gn")
    gn2 = gn/2
    np.testing.assert_almost_equal(
            gn2.getVariance(), 0.25, precision,
            err_msg="GaussianNoise gn/2 results in wrong variance")
    np.testing.assert_almost_equal(
            gn.getVariance(), 0.5, precision,
            err_msg="GaussianNoise 5*gn results in wrong variance for original gn")
    gn *= 3
    np.testing.assert_almost_equal(
            gn.getVariance(), 1.5, precision,
            err_msg="GaussianNoise gn*=3 results in wrong variance")
    gn /= 2
    np.testing.assert_almost_equal(
            gn.getVariance(), 0.75, precision,
            err_msg="GaussianNoise gn/=2 results in wrong variance")

    # Check starting with GaussianNoise()
    gn2 = galsim.GaussianNoise()
    gn2 = gn2.withVariance(9.)
    np.testing.assert_almost_equal(
            gn2.getVariance(), 9, precision,
            err_msg="GaussianNoise().withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            gn2.sigma, 3., precision,
            err_msg="GaussianNoise().withVariance results in wrong sigma")

    gn2 = galsim.GaussianNoise()
    gn2 = gn2.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            gn2.getVariance(), 4., precision,
            err_msg="GaussianNoise().withScaledVariance results in wrong variance")
    np.testing.assert_almost_equal(
            gn2.sigma, 2., precision,
            err_msg="GaussianNoise().withScaledVariance results in wrong sigma")

    # Check picklability
    check_pickle(gn, lambda x: (x.rng.serialize(), x.sigma))
    check_pickle(gn, drawNoise)
    check_pickle(gn)

    # Check copy, eq and ne
    gn = gn.withVariance(gSigma**2)
    gn2 = galsim.GaussianNoise(gn.rng.duplicate(), gSigma)
    gn3 = gn.copy()
    gn4 = gn.copy(rng=galsim.BaseDeviate(11))
    gn5 = galsim.GaussianNoise(gn.rng, 2.*gSigma)
    assert gn == gn2
    assert gn == gn3
    assert gn != gn4
    assert gn != gn5
    assert gn.rng.raw() == gn2.rng.raw()
    assert gn == gn2
    assert gn == gn3
    gn.rng.raw()
    assert gn != gn2
    assert gn == gn3


@timer
def test_variable_gaussian_noise():
    """Test VariableGaussian random number generator
    """
    # Make a checkerboard image with two values for the variance
    gSigma1 = 17.23
    gSigma2 = 28.55
    var_image = galsim.ImageD(galsim.BoundsI(0,9,0,9))
    coords = np.ogrid[0:10, 0:10]
    var_image.array[ (coords[0] + coords[1]) % 2 == 1 ] = gSigma1**2
    var_image.array[ (coords[0] + coords[1]) % 2 == 0 ] = gSigma2**2
    print('var_image.array = ',var_image.array)

    g = galsim.GaussianDeviate(testseed, sigma=1.)
    vgResult = np.empty((10,10))
    g.generate(vgResult)
    vgResult *= np.sqrt(var_image.array)

    # Test filling an image
    vgn = galsim.VariableGaussianNoise(galsim.BaseDeviate(testseed), var_image)
    testimage = galsim.ImageD(10,10)
    testimage.addNoise(vgn)
    np.testing.assert_array_almost_equal(
            testimage.array, vgResult, precision,
            err_msg="VariableGaussianNoise applied to Images does not reproduce expected sequence")

    # Test filling an image with Fortran ordering
    vgn.rng.seed(testseed)
    testimage = galsim.ImageD(np.zeros((10,10)).T)
    testimage.addNoise(vgn)
    np.testing.assert_array_almost_equal(
            testimage.array, vgResult, precision,
            err_msg="Wrong VariableGaussian noise generated for Fortran-ordered Image")

    # Check var_image property
    np.testing.assert_almost_equal(
            vgn.var_image.array, var_image.array, precision,
            err_msg="VariableGaussianNoise var_image returns wrong var_image")

    # Check that the noise model really does produce this variance.
    big_var_image = galsim.ImageD(galsim.BoundsI(0,2047,0,2047))
    big_coords = np.ogrid[0:2048, 0:2048]
    mask1 = (big_coords[0] + big_coords[1]) % 2 == 0
    mask2 = (big_coords[0] + big_coords[1]) % 2 == 1
    big_var_image.array[mask1] = gSigma1**2
    big_var_image.array[mask2] = gSigma2**2
    big_vgn = galsim.VariableGaussianNoise(galsim.BaseDeviate(testseed), big_var_image)

    big_im = galsim.Image(2048,2048,dtype=float)
    big_im.addNoise(big_vgn)
    var = np.var(big_im.array)
    print('variance = ',var)
    print('getVar = ',big_vgn.var_image.array.mean())
    np.testing.assert_almost_equal(
            var, big_vgn.var_image.array.mean(), 1,
            err_msg='Realized variance for VariableGaussianNoise did not match var_image')

    # Check realized variance in each mask
    print('rms1 = ',np.std(big_im.array[mask1]))
    print('rms2 = ',np.std(big_im.array[mask2]))
    np.testing.assert_almost_equal(np.std(big_im.array[mask1]), gSigma1, decimal=1)
    np.testing.assert_almost_equal(np.std(big_im.array[mask2]), gSigma2, decimal=1)

    # Check that VariableGaussianNoise adds to the image, not overwrites the image.
    gal = galsim.Exponential(half_light_radius=2.3, flux=1.e4)
    gal.drawImage(image=big_im)
    big_vgn.rng.seed(testseed)
    big_im.addNoise(big_vgn)
    gal.withFlux(-1.e4).drawImage(image=big_im, add_to_image=True)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(
            var, big_vgn.var_image.array.mean(), 1,
            err_msg='VariableGaussianNoise wrong when already an object drawn on the image')

    # Check picklability
    check_pickle(vgn, lambda x: (x.rng.serialize(), x.var_image))
    check_pickle(vgn, drawNoise)
    check_pickle(vgn)

    # Check copy, eq and ne
    vgn2 = galsim.VariableGaussianNoise(vgn.rng.duplicate(), var_image)
    vgn3 = vgn.copy()
    vgn4 = vgn.copy(rng=galsim.BaseDeviate(11))
    vgn5 = galsim.VariableGaussianNoise(vgn.rng, 2.*var_image)
    assert vgn == vgn2
    assert vgn == vgn3
    assert vgn != vgn4
    assert vgn != vgn5
    assert vgn.rng.raw() == vgn2.rng.raw()
    assert vgn == vgn2
    assert vgn == vgn3
    vgn.rng.raw()
    assert vgn != vgn2
    assert vgn == vgn3

    assert_raises(TypeError, vgn.applyTo, 23)
    assert_raises(ValueError, vgn.applyTo, galsim.ImageF(3,3))
    assert_raises(galsim.GalSimError, vgn.getVariance)
    assert_raises(galsim.GalSimError, vgn.withVariance, 23)
    assert_raises(galsim.GalSimError, vgn.withScaledVariance, 23)


@timer
def test_poisson_noise():
    """Test Poisson random number generator
    """
    pMean = 17
    p = galsim.PoissonDeviate(testseed, mean=pMean)
    pResult = np.empty((10,10))
    p.generate(pResult)
    noise = galsim.DeviateNoise(p)

    # Test filling an image
    noise.rng.seed(testseed)
    testimage = galsim.ImageI(10, 10)
    testimage.addNoise(galsim.DeviateNoise(p))
    np.testing.assert_array_equal(
            testimage.array, pResult,
            err_msg='Wrong poisson random number sequence generated when applied to image.')

    # The PoissonNoise version also subtracts off the mean value
    pn = galsim.PoissonNoise(galsim.BaseDeviate(testseed), sky_level=pMean)
    testimage.fill(0)
    testimage.addNoise(pn)
    np.testing.assert_array_equal(
            testimage.array, pResult-pMean,
            err_msg='Wrong poisson random number sequence generated using PoissonNoise')

    # Test filling a single-precision image
    pn.rng.seed(testseed)
    testimage = galsim.ImageF(10,10)
    testimage.addNoise(pn)
    np.testing.assert_array_almost_equal(
            testimage.array, pResult-pMean, precisionF,
            err_msg='Wrong Poisson random number sequence generated when applied to ImageF.')

    # Test filling an image with Fortran ordering
    pn.rng.seed(testseed)
    testimage = galsim.ImageD(10,10)
    testimage.addNoise(pn)
    np.testing.assert_array_almost_equal(
            testimage.array, pResult-pMean,
            err_msg="Wrong Poisson noise generated for Fortran-ordered Image")

    # Check PoissonNoise variance:
    np.testing.assert_almost_equal(
            pn.getVariance(), pMean, precision,
            err_msg="PoissonNoise getVariance returns wrong variance")
    np.testing.assert_almost_equal(
            pn.sky_level, pMean, precision,
            err_msg="PoissonNoise sky_level returns wrong value")

    # Check that the noise model really does produce this variance.
    big_im = galsim.Image(2048,2048,dtype=float)
    big_im.addNoise(pn)
    var = np.var(big_im.array)
    print('variance = ',var)
    print('getVar = ',pn.getVariance())
    np.testing.assert_almost_equal(
            var, pn.getVariance(), 1,
            err_msg='Realized variance for PoissonNoise did not match getVariance()')

    # Check that PoissonNoise adds to the image, not overwrites the image.
    gal = galsim.Exponential(half_light_radius=2.3, flux=0.3)
    # Note: in this case, flux/size^2 needs to be << sky_level or it will mess up the statistics.
    gal.drawImage(image=big_im)
    big_im.addNoise(pn)
    gal.withFlux(-0.3).drawImage(image=big_im, add_to_image=True)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(
            var, pn.getVariance(), 1,
            err_msg='PoissonNoise wrong when already an object drawn on the image')

    # Check withVariance
    pn = pn.withVariance(9.)
    np.testing.assert_almost_equal(
            pn.getVariance(), 9., precision,
            err_msg="PoissonNoise withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            pn.sky_level, 9., precision,
            err_msg="PoissonNoise withVariance results in wrong sky_level")

    # Check withScaledVariance
    pn = pn.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            pn.getVariance(), 36, precision,
            err_msg="PoissonNoise withScaledVariance results in wrong variance")
    np.testing.assert_almost_equal(
            pn.sky_level, 36., precision,
            err_msg="PoissonNoise withScaledVariance results in wrong sky_level")

    # Check arithmetic
    pn = pn.withVariance(0.5)
    pn2 = pn * 3
    np.testing.assert_almost_equal(
            pn2.getVariance(), 1.5, precision,
            err_msg="PoissonNoise pn*3 results in wrong variance")
    np.testing.assert_almost_equal(
            pn.getVariance(), 0.5, precision,
            err_msg="PoissonNoise pn*3 results in wrong variance for original pn")
    pn2 = 5 * pn
    np.testing.assert_almost_equal(
            pn2.getVariance(), 2.5, precision,
            err_msg="PoissonNoise 5*pn results in wrong variance")
    np.testing.assert_almost_equal(
            pn.getVariance(), 0.5, precision,
            err_msg="PoissonNoise 5*pn results in wrong variance for original pn")
    pn2 = pn/2
    np.testing.assert_almost_equal(
            pn2.getVariance(), 0.25, precision,
            err_msg="PoissonNoise pn/2 results in wrong variance")
    np.testing.assert_almost_equal(
            pn.getVariance(), 0.5, precision,
            err_msg="PoissonNoise 5*pn results in wrong variance for original pn")
    pn *= 3
    np.testing.assert_almost_equal(
            pn.getVariance(), 1.5, precision,
            err_msg="PoissonNoise pn*=3 results in wrong variance")
    pn /= 2
    np.testing.assert_almost_equal(
            pn.getVariance(), 0.75, precision,
            err_msg="PoissonNoise pn/=2 results in wrong variance")

    # Check starting with PoissonNoise()
    pn = galsim.PoissonNoise()
    pn = pn.withVariance(9.)
    np.testing.assert_almost_equal(
            pn.getVariance(), 9., precision,
            err_msg="PoissonNoise().withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            pn.sky_level, 9., precision,
            err_msg="PoissonNoise().withVariance results in wrong sky_level")
    pn = pn.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            pn.getVariance(), 36, precision,
            err_msg="PoissonNoise().withScaledVariance results in wrong variance")
    np.testing.assert_almost_equal(
            pn.sky_level, 36., precision,
            err_msg="PoissonNoise().withScaledVariance results in wrong sky_level")

    # Check picklability
    check_pickle(pn, lambda x: (x.rng.serialize(), x.sky_level))
    check_pickle(pn, drawNoise)
    check_pickle(pn)

    # Check copy, eq and ne
    pn = pn.withVariance(pMean)
    pn2 = galsim.PoissonNoise(pn.rng.duplicate(), pMean)
    pn3 = pn.copy()
    pn4 = pn.copy(rng=galsim.BaseDeviate(11))
    pn5 = galsim.PoissonNoise(pn.rng, 2*pMean)
    assert pn == pn2
    assert pn == pn3
    assert pn != pn4
    assert pn != pn5
    assert pn.rng.raw() == pn2.rng.raw()
    assert pn == pn2
    assert pn == pn3
    pn.rng.raw()
    assert pn != pn2
    assert pn == pn3


@timer
def test_ccdnoise():
    """Test CCD Noise generator
    """
    # Start with some regression tests where we have known values that we expect to generate:

    types = (np.int16, np.int32, np.float32, np.float64)
    typestrings = ("S", "I", "F", "D")

    testseed = 1000
    gain = 3.
    read_noise = 5.
    sky = 50

    # Tabulated results for the above settings and testseed value.
    cResultS = np.array([[44, 47], [50, 49]], dtype=np.int16)
    cResultI = np.array([[44, 47], [50, 49]], dtype=np.int32)
    cResultF = np.array([[44.45332718, 47.79725266], [50.67744064, 49.58272934]], dtype=np.float32)
    cResultD = np.array([[44.453328440057618, 47.797254142519577],
                        [50.677442088335162, 49.582730949808081]],dtype=np.float64)

    for i in range(4):
        prec = eval("precision"+typestrings[i])
        cResult = eval("cResult"+typestrings[i])

        rng = galsim.BaseDeviate(testseed)
        ccdnoise = galsim.CCDNoise(rng, gain=gain, read_noise=read_noise)
        testImage = galsim.Image((np.zeros((2, 2))+sky).astype(types[i]))
        ccdnoise.applyTo(testImage)
        np.testing.assert_array_almost_equal(
                testImage.array, cResult, prec,
                err_msg="Wrong CCD noise random sequence generated for Image"+typestrings[i]+".")

        # Check that reseeding the rng reseeds the internal deviate in CCDNoise
        rng.seed(testseed)
        testImage.fill(sky)
        ccdnoise.applyTo(testImage)
        np.testing.assert_array_almost_equal(
                testImage.array, cResult, prec,
                err_msg="Wrong CCD noise random sequence generated for Image"+typestrings[i]+
                " after seed")

        # Check using addNoise
        rng.seed(testseed)
        testImage.fill(sky)
        testImage.addNoise(ccdnoise)
        np.testing.assert_array_almost_equal(
                testImage.array, cResult, prec,
                err_msg="Wrong CCD noise random sequence generated for Image"+typestrings[i]+
                " using addNoise")

        # Test filling an image with Fortran ordering
        rng.seed(testseed)
        testImageF = galsim.Image(np.zeros((2, 2)).T, dtype=types[i])
        testImageF.fill(sky)
        testImageF.addNoise(ccdnoise)
        np.testing.assert_array_almost_equal(
                testImageF.array, cResult, prec,
                err_msg="Wrong CCD noise generated for Fortran-ordered Image"+typestrings[i])

        # Now include sky_level in ccdnoise
        rng.seed(testseed)
        ccdnoise = galsim.CCDNoise(rng, sky_level=sky, gain=gain, read_noise=read_noise)
        testImage.fill(0)
        ccdnoise.applyTo(testImage)
        np.testing.assert_array_almost_equal(
                testImage.array, cResult-sky, prec,
                err_msg="Wrong CCD noise random sequence generated for Image"+typestrings[i]+
                " with sky_level included in noise")

        rng.seed(testseed)
        testImage.fill(0)
        testImage.addNoise(ccdnoise)
        np.testing.assert_array_almost_equal(
                testImage.array, cResult-sky, prec,
                err_msg="Wrong CCD noise random sequence generated for Image"+typestrings[i]+
                " using addNoise with sky_level included in noise")

    # Check CCDNoise variance:
    var1 = sky/gain + (read_noise/gain)**2
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), var1, precision,
            err_msg="CCDNoise getVariance returns wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, sky, precision,
            err_msg="CCDNoise sky_level returns wrong value")
    np.testing.assert_almost_equal(
            ccdnoise.gain, gain, precision,
            err_msg="CCDNoise gain returns wrong value")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, read_noise, precision,
            err_msg="CCDNoise read_noise returns wrong value")

    # Check that the noise model really does produce this variance.
    # NB. If default float32 is used here, older versions of numpy will compute the variance
    # in single precision, and with 2048^2 values, the final answer comes out significantly
    # wrong (19.33 instead of 19.42, which gets compared to the nominal value of 19.44).
    big_im = galsim.Image(2048,2048,dtype=float)
    big_im.addNoise(ccdnoise)
    var = np.var(big_im.array)
    print('variance = ',var)
    print('getVar = ',ccdnoise.getVariance())
    np.testing.assert_almost_equal(
            var, ccdnoise.getVariance(), 1,
            err_msg='Realized variance for CCDNoise did not match getVariance()')

    # Check that CCDNoise adds to the image, not overwrites the image.
    gal = galsim.Exponential(half_light_radius=2.3, flux=0.3)
    # Note: again, flux/size^2 needs to be << sky_level or it will mess up the statistics.
    gal.drawImage(image=big_im)
    big_im.addNoise(ccdnoise)
    gal.withFlux(-0.3).drawImage(image=big_im, add_to_image=True)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(
            var, ccdnoise.getVariance(), 1,
            err_msg='CCDNoise wrong when already an object drawn on the image')

    # Check using a non-integer sky level which does some slightly different calculations.
    rng.seed(testseed)
    big_im_int = galsim.Image(2048,2048,dtype=int)
    ccdnoise = galsim.CCDNoise(rng, sky_level=34.42, gain=1.6, read_noise=11.2)
    big_im_int.fill(0)
    big_im_int.addNoise(ccdnoise)
    var = np.var(big_im_int.array)
    np.testing.assert_almost_equal(var/ccdnoise.getVariance(), 1., decimal=2,
                                   err_msg='CCDNoise wrong when sky_level is not an integer')

    # Using gain=0 means the read_noise is in ADU, not e-
    rng.seed(testseed)
    ccdnoise = galsim.CCDNoise(rng, gain=0., read_noise=read_noise)
    var2 = read_noise**2
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), var2, precision,
            err_msg="CCDNoise getVariance returns wrong variance with gain=0")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, 0., precision,
            err_msg="CCDNoise sky_level returns wrong value with gain=0")
    np.testing.assert_almost_equal(
            ccdnoise.gain, 0., precision,
            err_msg="CCDNoise gain returns wrong value with gain=0")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, read_noise, precision,
            err_msg="CCDNoise read_noise returns wrong value with gain=0")
    big_im.fill(0)
    big_im.addNoise(ccdnoise)
    var = np.var(big_im.array)
    np.testing.assert_almost_equal(var, ccdnoise.getVariance(), 1,
                                   err_msg='CCDNoise wrong when gain=0')

    # Check withVariance
    ccdnoise = galsim.CCDNoise(rng, sky_level=sky, gain=gain, read_noise=read_noise)
    ccdnoise = ccdnoise.withVariance(9.)
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 9., precision,
            err_msg="CCDNoise withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, (9./var1)*sky, precision,
            err_msg="CCDNoise withVariance results in wrong sky_level")
    np.testing.assert_almost_equal(
            ccdnoise.gain, gain, precision,
            err_msg="CCDNoise withVariance results in wrong gain")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, np.sqrt(9./var1) * read_noise, precision,
            err_msg="CCDNoise withVariance results in wrong ReadNoise")

    # Check withScaledVariance
    ccdnoise = ccdnoise.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 36., precision,
            err_msg="CCDNoise withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, (36./var1)*sky, precision,
            err_msg="CCDNoise withVariance results in wrong sky_level")
    np.testing.assert_almost_equal(
            ccdnoise.gain, gain, precision,
            err_msg="CCDNoise withVariance results in wrong gain")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, np.sqrt(36./var1) * read_noise, precision,
            err_msg="CCDNoise withVariance results in wrong ReadNoise")

    # Check arithmetic
    ccdnoise = ccdnoise.withVariance(0.5)
    ccdnoise2 = ccdnoise * 3
    np.testing.assert_almost_equal(
            ccdnoise2.getVariance(), 1.5, precision,
            err_msg="CCDNoise ccdnoise*3 results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 0.5, precision,
            err_msg="CCDNoise ccdnoise*3 results in wrong variance for original ccdnoise")
    ccdnoise2 = 5 * ccdnoise
    np.testing.assert_almost_equal(
            ccdnoise2.getVariance(), 2.5, precision,
            err_msg="CCDNoise 5*ccdnoise results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 0.5, precision,
            err_msg="CCDNoise 5*ccdnoise results in wrong variance for original ccdnoise")
    ccdnoise2 = ccdnoise/2
    np.testing.assert_almost_equal(
            ccdnoise2.getVariance(), 0.25, precision,
            err_msg="CCDNoise ccdnoise/2 results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 0.5, precision,
            err_msg="CCDNoise 5*ccdnoise results in wrong variance for original ccdnoise")
    ccdnoise *= 3
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 1.5, precision,
            err_msg="CCDNoise ccdnoise*=3 results in wrong variance")
    ccdnoise /= 2
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 0.75, precision,
            err_msg="CCDNoise ccdnoise/=2 results in wrong variance")

    # Check starting with CCDNoise()
    ccdnoise = galsim.CCDNoise()
    ccdnoise = ccdnoise.withVariance(9.)
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 9., precision,
            err_msg="CCDNoise().withVariance results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, 9., precision,
            err_msg="CCDNoise().withVariance results in wrong sky_level")
    np.testing.assert_almost_equal(
            ccdnoise.gain, 1., precision,
            err_msg="CCDNoise().withVariance results in wrong gain")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, 0., precision,
            err_msg="CCDNoise().withVariance results in wrong ReadNoise")
    ccdnoise = ccdnoise.withScaledVariance(4.)
    np.testing.assert_almost_equal(
            ccdnoise.getVariance(), 36., precision,
            err_msg="CCDNoise().withScaledVariance results in wrong variance")
    np.testing.assert_almost_equal(
            ccdnoise.sky_level, 36., precision,
            err_msg="CCDNoise().withScaledVariance results in wrong sky_level")
    np.testing.assert_almost_equal(
            ccdnoise.gain, 1., precision,
            err_msg="CCDNoise().withScaledVariance results in wrong gain")
    np.testing.assert_almost_equal(
            ccdnoise.read_noise, 0., precision,
            err_msg="CCDNoise().withScaledVariance results in wrong ReadNoise")

    # Check picklability
    check_pickle(ccdnoise, lambda x: (x.rng.serialize(), x.sky_level, x.gain, x.read_noise))
    check_pickle(ccdnoise, drawNoise)
    check_pickle(ccdnoise)

    # Check copy, eq and ne
    ccdnoise = galsim.CCDNoise(rng, sky, gain, read_noise)
    ccdnoise2 = galsim.CCDNoise(ccdnoise.rng.duplicate(), gain=gain, read_noise=read_noise,
                                sky_level=sky)
    ccdnoise3 = ccdnoise.copy()
    ccdnoise4 = ccdnoise.copy(rng=galsim.BaseDeviate(11))
    ccdnoise5 = galsim.CCDNoise(ccdnoise.rng, gain=2*gain, read_noise=read_noise, sky_level=sky)
    ccdnoise6 = galsim.CCDNoise(ccdnoise.rng, gain=gain, read_noise=2*read_noise, sky_level=sky)
    ccdnoise7 = galsim.CCDNoise(ccdnoise.rng, gain=gain, read_noise=read_noise, sky_level=2*sky)
    assert ccdnoise == ccdnoise2
    assert ccdnoise == ccdnoise3
    assert ccdnoise != ccdnoise4
    assert ccdnoise != ccdnoise5
    assert ccdnoise != ccdnoise6
    assert ccdnoise != ccdnoise7
    assert ccdnoise.rng.raw() == ccdnoise2.rng.raw()
    assert ccdnoise == ccdnoise2
    assert ccdnoise == ccdnoise3
    ccdnoise.rng.raw()
    assert ccdnoise != ccdnoise2
    assert ccdnoise == ccdnoise3


@timer
def test_addnoisesnr():
    """Test that addNoiseSNR is behaving sensibly.
    """
    # Rather than reproducing the S/N calculation in addNoiseSNR(), we'll just check for
    # self-consistency of the behavior with / without flux preservation.
    # Begin by making some object that we draw into an Image.
    gal_sigma = 3.7
    pix_scale = 0.6
    test_snr = 73.
    gauss = galsim.Gaussian(sigma=gal_sigma)
    im = gauss.drawImage(scale=pix_scale, dtype=np.float64)

    # Now make the noise object to use.
    # Use a default-constructed rng (i.e. rng=None) since we had initially had trouble
    # with that.  And use the duplicate feature to get a second copy of this rng.
    gn = galsim.GaussianNoise()
    rng2 = gn.rng.duplicate()

    # Try addNoiseSNR with preserve_flux=True, so the RNG needs a different variance.
    # Check what variance was added for this SNR, and that the RNG still has its original variance
    # after this call.
    var_out = im.addNoiseSNR(gn, test_snr, preserve_flux=True)
    assert gn.getVariance()==1.0
    max_val = im.array.max()

    # Now apply addNoiseSNR to another (clean) image with preserve_flux=False, so we use the noise
    # variance in the original RNG, i.e., 1.  Check that the returned variance is 1, and that the
    # value of the maximum pixel (presumably the peak of the galaxy light profile) is scaled as we
    # expect for this SNR.
    im2 = gauss.drawImage(scale=pix_scale, dtype=np.float64)
    gn2 = galsim.GaussianNoise(rng=rng2)
    var_out2 = im2.addNoiseSNR(gn2, test_snr, preserve_flux=False)
    assert var_out2==1.0
    expect_max_val2 = max_val*np.sqrt(var_out2/var_out)
    np.testing.assert_almost_equal(
            im2.array.max(), expect_max_val2, decimal=8,
            err_msg='addNoiseSNR with preserve_flux = True and False give inconsistent results')


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from .input import InputLoader, GetInputObj, RegisterInputType, GetNumInputObj
from .util import GetRNG, get_cls_params
from .value import GetAllParams, SetDefaultIndex
from .gsobject import RegisterObjectType
from ..gsparams import GSParams
from ..errors import GalSimConfigError
from ..real import RealGalaxyCatalog, RealGalaxy, ChromaticRealGalaxy

# This file adds input type real_catalog and gsobject types RealGalaxy and RealGalaxyOriginal.

# The RealGalaxyCatalog doesn't need anything special other than registration as a valid
# input type.
RegisterInputType('real_catalog', InputLoader(RealGalaxyCatalog, takes_logger=True))

# There are two gsobject types that are coupled to this: RealGalaxy and RealGalaxyOriginal.

def _BuildRealGalaxy(config, base, ignore, gsparams, logger, param_name='RealGalaxy'):
    """Build a RealGalaxy from the real_catalog input item.
    """
    real_cat = GetInputObj('real_catalog', config, base, param_name)

    # Special: if index is Sequence or Random, and max isn't set, set it to nobjects-1.
    # But not if they specify 'id' or have 'random=True', which overrides that.
    if 'id' not in config:
        if 'random' not in config:
            SetDefaultIndex(config, real_cat.nobjects)
        else:
            if not config['random']:
                SetDefaultIndex(config, real_cat.nobjects)
                # Need to do this to avoid being caught by the GetAllParams() call, which will flag
                # it if it has 'index' and 'random' set (but 'random' is False, so really it's OK).
                del config['random']

    req, opt, single, _ = get_cls_params(RealGalaxy)
    kwargs, safe = GetAllParams(config, base, req, opt, single, ignore = ignore + ['num'])
    if gsparams: kwargs['gsparams'] = GSParams(**gsparams)

    kwargs['rng'] = GetRNG(config, base, logger, param_name)

    if 'index' in kwargs:
        index = kwargs['index']
        if index >= real_cat.nobjects or index < 0:
            raise GalSimConfigError(
                "index=%s has gone past the number of entries in the RealGalaxyCatalog"%index)

    kwargs['real_galaxy_catalog'] = real_cat
    logger.debug('obj %d: %s kwargs = %s',base.get('obj_num',0),param_name,kwargs)

    gal = RealGalaxy(**kwargs)

    return gal, safe


def _BuildRealGalaxyOriginal(config, base, ignore, gsparams, logger):
    """Return the original image from a RealGalaxy using the real_catalog input item.
    """
    gal, safe = _BuildRealGalaxy(config, base, ignore, gsparams, logger,
                                   param_name='RealGalaxyOriginal')
    return gal.original_gal, safe

def _BuildChromaticRealGalaxy(config, base, ignore, gsparams, logger):
    """Build a ChromaticRealGalaxy from several real_catalog input items.
    """
    param_name = 'ChromaticRealGalaxy'
    ncats = GetNumInputObj('real_catalog', base)
    real_cats = [ GetInputObj('real_catalog', config, base, param_name, n)
                  for n in range(ncats)]

    # Special: if index is Sequence or Random, and max isn't set, set it to nobjects-1.
    # But not if they specify 'id' or have 'random=True', which overrides that.
    if 'id' not in config:
        if 'random' not in config:
            SetDefaultIndex(config, real_cats[0].nobjects)
        else:
            if not config['random']:
                SetDefaultIndex(config, real_cats[0].nobjects)
                # Need to do this to avoid being caught by the GetAllParams() call, which will flag
                # it if it has 'index' and 'random' set (but 'random' is False, so really it's OK).
                del config['random']

    req, opt, single, _ = get_cls_params(ChromaticRealGalaxy)

    kwargs, safe = GetAllParams(config, base, req, opt, single, ignore)
    if gsparams: kwargs['gsparams'] = GSParams(**gsparams)

    kwargs['rng'] = GetRNG(config, base, logger, param_name)

    if 'index' in kwargs:
        index = kwargs['index']
        if index >= real_cats[0].nobjects or index < 0:
            raise GalSimConfigError(
                "index=%s has gone past the number of entries in the RealGalaxyCatalog"%index)

    kwargs['real_galaxy_catalogs'] = real_cats
    logger.debug('obj %d: %s kwargs = %s',base.get('obj_num',0),param_name,kwargs)

    gal = ChromaticRealGalaxy(**kwargs)

    return gal, safe


# Register these as valid gsobject types
RegisterObjectType('RealGalaxy', _BuildRealGalaxy, input_type='real_catalog')
RegisterObjectType('RealGalaxyOriginal', _BuildRealGalaxyOriginal, input_type='real_catalog')
RegisterObjectType('ChromaticRealGalaxy', _BuildChromaticRealGalaxy, input_type='real_catalog')



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__)
imgdir = os.path.join(path, "SBProfile_comparison_images") # Directory containing the reference
                                                           # images.


@timer
def test_gaussian():
    """Test the generation of a specific Gaussian profile against a known result.
    """
    savedImg = galsim.fits.read(os.path.join(imgdir, "gauss_1.fits"))
    savedImg.setCenter(0,0)
    dx = 0.2
    myImg = galsim.ImageF(savedImg.bounds, scale=dx)
    myImg.setCenter(0,0)

    gauss = galsim.Gaussian(flux=1, sigma=1)
    # Reference images were made with old centering, which is equivalent to use_true_center=False.
    myImg = gauss.drawImage(myImg, scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Gaussian disagrees with expected result")
    np.testing.assert_almost_equal(
            myImg.array.sum(dtype=float) *dx**2, myImg.added_flux, 5,
            err_msg="Gaussian profile GSObject::draw returned wrong added_flux")

    # Check a non-square image
    print(myImg.bounds)
    recImg = galsim.ImageF(45,66)
    recImg.setCenter(0,0)
    recImg = gauss.drawImage(recImg, scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            recImg[savedImg.bounds].array, savedImg.array, 5,
            err_msg="Drawing Gaussian on non-square image disagrees with expected result")
    np.testing.assert_almost_equal(
            recImg.array.sum(dtype=float) *dx**2, recImg.added_flux, 5,
            err_msg="Gaussian profile GSObject::draw on non-square image returned wrong added_flux")

    # Check with default_params
    gauss = galsim.Gaussian(flux=1, sigma=1, gsparams=default_params)
    gauss.drawImage(myImg,scale=0.2, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Gaussian with default_params disagrees with expected result")
    gauss = galsim.Gaussian(flux=1, sigma=1, gsparams=galsim.GSParams())
    gauss.drawImage(myImg,scale=0.2, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Gaussian with GSParams() disagrees with expected result")

    # Use non-unity values.
    gauss = galsim.Gaussian(flux=1.7, sigma=2.3)
    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    gauss2 = galsim.Gaussian(flux=1.7, sigma=2.3, gsparams=gsp)
    assert gauss2 != gauss
    assert gauss2 == gauss.withGSParams(gsp)
    assert gauss2 == gauss.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    assert gauss2 == gauss.withGSParams(xvalue_accuracy=1.e-8).withGSParams(kvalue_accuracy=1.e-8)
    assert gauss2 == gauss.withGSParams(galsim.GSParams(xvalue_accuracy=1.e-8),
                                        kvalue_accuracy=1.e-8)
    assert gauss2 == gauss.withGSParams(gsp).withGSParams(kvalue_accuracy=1.e-8)
    assert gauss2 == gauss.withGSParams(galsim.GSParams(xvalue_accuracy=1.e-8)).withGSParams(
                                            kvalue_accuracy=1.e-8)
    check_basic(gauss, "Gaussian")

    # Check invalid parameters
    assert_raises(TypeError, gauss.withGSParams, xvalue_threshold=1.e-8)
    assert_raises(TypeError, gauss.withGSParams, xvalue_accuracy=1.e-8, kvalue=1.e-8)

    # Test photon shooting.
    do_shoot(gauss,myImg,"Gaussian")

    # Test kvalues
    do_kvalue(gauss,myImg,"Gaussian")

    # Check picklability
    check_pickle(galsim.GSParams())  # Check GSParams explicitly here too.
    check_pickle(galsim.GSParams(
        minimum_fft_size = 12,
        maximum_fft_size = 40,
        folding_threshold = 1.e-1,
        maxk_threshold = 2.e-1,
        kvalue_accuracy = 3.e-1,
        xvalue_accuracy = 4.e-1,
        shoot_accuracy = 5.e-1,
        realspace_relerr = 6.e-1,
        realspace_abserr = 7.e-1,
        integration_relerr = 8.e-1,
        integration_abserr = 9.e-1))
    check_pickle(gauss, lambda x: x.drawImage(method='no_pixel'))
    check_pickle(gauss)

    # Should raise an exception if >=2 radii are provided.
    assert_raises(TypeError, galsim.Gaussian, sigma=3, half_light_radius=1, fwhm=2)
    assert_raises(TypeError, galsim.Gaussian, half_light_radius=1, fwhm=2)
    assert_raises(TypeError, galsim.Gaussian, sigma=3, fwhm=2)
    assert_raises(TypeError, galsim.Gaussian, sigma=3, half_light_radius=1)
    # Or none.
    assert_raises(TypeError, galsim.Gaussian)

    # Finally, test the noise property for things that don't have any noise set.
    assert gauss.noise is None
    # And accessing the attribute from the class should indicate that it is a lazyproperty
    assert 'lazy_property' in str(galsim.GSObject._noise)

    # And check that trying to use GSObject directly is an error.
    assert_raises(NotImplementedError, galsim.GSObject)


@timer
def test_gaussian_properties():
    """Test some basic properties of the Gaussian profile.
    """
    test_flux = 17.9
    test_sigma = 1.8
    gauss = galsim.Gaussian(flux=test_flux, sigma=test_sigma)
    # Check that we are centered on (0, 0)
    cen = galsim.PositionD(0, 0)
    np.testing.assert_equal(gauss.centroid, cen)
    # Check Fourier properties
    np.testing.assert_almost_equal(gauss.maxk, 3.7169221888498383 / test_sigma)
    np.testing.assert_almost_equal(gauss.stepk, 0.533644625664 / test_sigma)
    np.testing.assert_almost_equal(gauss.kValue(cen), (1+0j) * test_flux)
    np.testing.assert_almost_equal(gauss.flux, test_flux)
    import math
    np.testing.assert_almost_equal(gauss.xValue(cen), 1./(2.*math.pi) * test_flux / test_sigma**2)
    np.testing.assert_almost_equal(gauss.xValue(cen), gauss.max_sb)
    # Check input flux vs output flux
    for inFlux in np.logspace(-2, 2, 10):
        gauss = galsim.Gaussian(flux=inFlux, sigma=test_sigma)
        outFlux = gauss.flux
        np.testing.assert_almost_equal(outFlux, inFlux)

    # Check some valid and invalid ways to pass arguments to xValue
    # Same code applies to kValue and others, so just do this one.
    assert gauss.xValue(cen.x, cen.y) == gauss.xValue(cen)
    assert gauss.xValue(x=cen.x, y=cen.y) == gauss.xValue(cen)
    assert gauss.xValue( (cen.x, cen.y) ) == gauss.xValue(cen)
    assert_raises(TypeError, gauss.xValue, cen.x)
    assert_raises(TypeError, gauss.xValue, x=cen.x)
    assert_raises(TypeError, gauss.xValue, cen.x, y=cen.y)
    assert_raises(TypeError, gauss.xValue, dx=cen.x, dy=cen.y)
    assert_raises(TypeError, gauss.xValue, dx=cen.x, y=cen.y)
    assert_raises(TypeError, gauss.xValue, x=cen.x, dy=cen.y)
    assert_raises(TypeError, gauss.xValue, cen.x, cen.y, cen.y)
    assert_raises(TypeError, gauss.xValue, cen.x, cen.y, invalid=True)
    assert_raises(TypeError, gauss.xValue, pos=cen)

    # Check that stepk and maxk scale correctly with radius
    gauss2 = galsim.Gaussian(flux=test_flux, sigma=5*test_sigma)
    np.testing.assert_almost_equal(gauss2.maxk, gauss.maxk/5)
    np.testing.assert_almost_equal(gauss2.stepk, gauss.stepk/5)


@timer
def test_gaussian_radii():
    """Test initialization of Gaussian with different types of radius specification.
    """
    import math
    test_hlr = 1.7
    test_fwhm = 1.8
    test_sigma = 1.9

    # Test constructor using half-light-radius:
    test_gal = galsim.Gaussian(flux=1., half_light_radius=test_hlr)
    hlr_sum = radial_integrate(test_gal, 0., test_hlr)
    print('hlr_sum = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=4,
            err_msg="Error in Gaussian constructor with half-light radius")

    # test that fwhm provides correct FWHM
    got_fwhm = test_gal.fwhm
    test_fwhm_ratio = (test_gal.xValue(galsim.PositionD(.5 * got_fwhm, 0.)) /
                       test_gal.xValue(galsim.PositionD(0., 0.)))
    print('fwhm ratio = ', test_fwhm_ratio)
    np.testing.assert_almost_equal(
            test_fwhm_ratio, 0.5, decimal=4,
            err_msg="Error in FWHM for Gaussian initialized with half-light radius")

    # test that sigma provides correct sigma
    got_sigma = test_gal.sigma
    test_sigma_ratio = (test_gal.xValue(galsim.PositionD(got_sigma, 0.)) /
                        test_gal.xValue(galsim.PositionD(0., 0.)))
    print('sigma ratio = ', test_sigma_ratio)
    np.testing.assert_almost_equal(
            test_sigma_ratio, math.exp(-0.5), decimal=4,
            err_msg="Error in sigma for Gaussian initialized with half-light radius")

    # Test constructor using sigma:
    test_gal = galsim.Gaussian(flux=1., sigma=test_sigma)
    center = test_gal.xValue(galsim.PositionD(0,0))
    ratio = test_gal.xValue(galsim.PositionD(test_sigma,0)) / center
    print('sigma ratio = ',ratio)
    np.testing.assert_almost_equal(
            ratio, np.exp(-0.5), decimal=4,
            err_msg="Error in Gaussian constructor with sigma")

    # then test that image indeed has the correct HLR properties when radially integrated
    got_hlr = test_gal.half_light_radius
    hlr_sum = radial_integrate(test_gal, 0., got_hlr)
    print('hlr_sum (profile initialized with sigma) = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=4,
            err_msg="Error in half light radius for Gaussian initialized with sigma.")

    # test that fwhm provides correct FWHM
    got_fwhm = test_gal.fwhm
    test_fwhm_ratio = (test_gal.xValue(galsim.PositionD(.5 * got_fwhm, 0.)) /
                       test_gal.xValue(galsim.PositionD(0., 0.)))
    print('fwhm ratio = ', test_fwhm_ratio)
    np.testing.assert_almost_equal(
            test_fwhm_ratio, 0.5, decimal=4,
            err_msg="Error in FWHM for Gaussian initialized with sigma.")

    # Test constructor using FWHM:
    test_gal = galsim.Gaussian(flux=1., fwhm=test_fwhm)
    center = test_gal.xValue(galsim.PositionD(0,0))
    ratio = test_gal.xValue(galsim.PositionD(test_fwhm/2.,0)) / center
    print('fwhm ratio = ',ratio)
    np.testing.assert_almost_equal(
            ratio, 0.5, decimal=4,
            err_msg="Error in Gaussian constructor with fwhm")

    # then test that image indeed has the correct HLR properties when radially integrated
    got_hlr = test_gal.half_light_radius
    hlr_sum = radial_integrate(test_gal, 0., got_hlr)
    print('hlr_sum (profile initialized with fwhm) = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=4,
            err_msg="Error in half light radius for Gaussian initialized with FWHM.")

    # test that sigma provides correct sigma
    got_sigma = test_gal.sigma
    test_sigma_ratio = (test_gal.xValue(galsim.PositionD(got_sigma, 0.)) /
                        test_gal.xValue(galsim.PositionD(0., 0.)))
    print('sigma ratio = ', test_sigma_ratio)
    np.testing.assert_almost_equal(
            test_sigma_ratio, math.exp(-0.5), decimal=4,
            err_msg="Error in sigma for Gaussian initialized with FWHM.")

    # Check that the properties don't work after modifying the original.
    # Note: I test all the modifiers here.  For the rest of the profile types, I'll
    # just confirm that it is true of shear.  I don't think that has any chance
    # of missing anything.
    test_gal_flux1 = test_gal * 3.
    assert_raises(AttributeError, getattr, test_gal_flux1, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_flux1, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_flux1, "sigma")

    test_gal_flux2 = test_gal.withScaledFlux(3.)
    assert_raises(AttributeError, getattr, test_gal_flux2, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_flux2, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_flux2, "sigma")

    test_gal_shear = test_gal.shear(g1=0.3, g2=0.1)
    assert_raises(AttributeError, getattr, test_gal_shear, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_shear, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_shear, "sigma")

    test_gal_rot = test_gal.rotate(theta = 0.5 * galsim.radians)
    assert_raises(AttributeError, getattr, test_gal_rot, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_rot, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_rot, "sigma")

    test_gal_shift = test_gal.shift(dx=0.11, dy=0.04)
    assert_raises(AttributeError, getattr, test_gal_shift, "fwhm")
    assert_raises(AttributeError, getattr, test_gal_shift, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_shift, "sigma")


@timer
def test_gaussian_flux_scaling():
    """Test flux scaling for Gaussian.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_flux = 17.9
    test_sigma = 1.8

    # init with sigma and flux only (should be ok given last tests)
    obj = galsim.Gaussian(sigma=test_sigma, flux=test_flux)
    obj *= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __imul__.")
    obj = galsim.Gaussian(sigma=test_sigma, flux=test_flux)
    obj /= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __idiv__.")
    obj = galsim.Gaussian(sigma=test_sigma, flux=test_flux)
    obj2 = obj * 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (result).")
    obj = galsim.Gaussian(sigma=test_sigma, flux=test_flux)
    obj2 = 2. * obj
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (result).")
    obj = galsim.Gaussian(sigma=test_sigma, flux=test_flux)
    obj2 = obj / 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (result).")


@timer
def test_gaussian_shoot():
    """Test Gaussian with photon shooting.  Particularly the flux of the final image.
    """
    rng = galsim.BaseDeviate(1234)
    obj = galsim.Gaussian(fwhm=3.5, flux=1.e4)
    im = galsim.Image(100,100, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng.duplicate())
    assert photons2 == photons, "Gaussian makePhot not equivalent to drawPhot"

    # Can treat the profile as a convolution of a delta function and put it in a photon_ops list.
    delta = galsim.DeltaFunction(flux=1.e4)
    psf = galsim.Gaussian(fwhm=3.5)
    photons3 = delta.makePhot(poisson_flux=False, rng=rng.duplicate(), photon_ops=[psf])
    assert photons3 == photons, "Using Gaussian in photon_ops not equivalent to drawPhot"


@timer
def test_ne():
    """Test base.py GSObjects for not-equals."""
    # Define some universal gsps
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)

    # Gaussian.  Params include sigma, fwhm, half_light_radius, flux, and gsparams.
    # The following should all test unequal:
    gals = [galsim.Gaussian(sigma=1.0),
            galsim.Gaussian(sigma=1.1),
            galsim.Gaussian(fwhm=1.0),
            galsim.Gaussian(half_light_radius=1.0),
            galsim.Gaussian(half_light_radius=1.1),
            galsim.Gaussian(sigma=1.2, flux=1.0),
            galsim.Gaussian(sigma=1.2, flux=1.1),
            galsim.Gaussian(sigma=1.2, gsparams=gsp)]
    # Check that setifying doesn't remove any duplicate items.
    check_all_diff(gals)


@timer
def test_accurate_shift():
    """Test that shifted Gaussian looks the same with real and fourier drawing.
    """
    # This is in response to issue #1231

    gal = galsim.Gaussian(sigma=1.).shift([5,5]).withFlux(200)

    real_im = gal.drawImage(nx=128, ny=128, scale=0.2, method='real_space')
    fft_im = gal.drawImage(nx=128, ny=128, scale=0.2, method='fft')
    print('max abs diff = ',np.max(np.abs(real_im.array - fft_im.array)))
    np.testing.assert_allclose(fft_im.array, real_im.array, rtol=1.e-7, atol=1.e-7)

    # In double precision it's almost perfect.
    real_im = gal.drawImage(nx=128, ny=128, scale=0.2, method='real_space', dtype=float)
    fft_im = gal.drawImage(nx=128, ny=128, scale=0.2, method='fft', dtype=float)
    print('max abs diff = ',np.max(np.abs(real_im.array - fft_im.array)))
    np.testing.assert_allclose(fft_im.array, real_im.array, rtol=1.e-14, atol=1.e-14)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import sys
import logging
from unittest import mock
import galsim
import galsim.main  # Not imported automatically
import galsim.__main__
from galsim_test_helpers import *

# This file tests the galsim executable.
# Note: Most of the real functional tests are elsewhere.  E.g. test_config*
#       This really just tests the bits related specifically to the functions in galsim/main.py.

@timer
def test_args():
    """Test the argument parsing of galsim executable.
    """
    # no args or -h prints description and exits
    print('This should print the description...')  # Can only check by inspection, not assert
    with assert_raises(SystemExit):
        args = galsim.main.parse_args([])
    print('This should print the description...')
    with assert_raises(SystemExit):
        args = galsim.main.parse_args(['-h'])

    # --version prints version and exits
    print('This should print the version...')
    with assert_raises(SystemExit):
        args = galsim.main.parse_args(['--version'])

    # Any other options, but missing config file, also prints description and exits
    print('This should print the description...')
    with assert_raises(SystemExit):
        args = galsim.main.parse_args(['-v','3','-n','5'])

    # Normal operation requires a config_file parameter
    config_file = 'test.yaml'
    args = galsim.main.parse_args([config_file])
    assert args.config_file == config_file
    assert args.verbosity == 1
    assert args.log_file is None
    assert args.log_format == '%(message)s'
    assert args.file_type is None
    assert args.module is None
    assert args.profile is False
    assert args.njobs == 1
    assert args.job == 1
    assert args.except_abort is False
    assert args.version is False
    assert args.variables == []

    # Check setting each of those to be something else
    args = galsim.main.parse_args([config_file, '-v', '3'])
    assert args.config_file == config_file
    assert args.verbosity == 3

    args = galsim.main.parse_args([config_file, '-l', 'test.log'])
    assert args.config_file == config_file
    assert args.log_file == 'test.log'

    args = galsim.main.parse_args([config_file, '--log_format', '%(asctime)s - %(name)s - %(levelname)s - %(message)s'])
    assert args.config_file == config_file
    assert args.log_format == '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    args = galsim.main.parse_args([config_file, '-f', 'yaml'])
    assert args.config_file == config_file
    assert args.file_type == 'yaml'
    args = galsim.main.parse_args([config_file, '-f', 'json'])
    assert args.config_file == config_file
    assert args.file_type == 'json'

    args = galsim.main.parse_args([config_file, '-m', 'des'])
    assert args.config_file == config_file
    assert args.module == ['des']
    args = galsim.main.parse_args([config_file, '-m', 'des', '-m', 'galsim_extra'])
    assert args.config_file == config_file
    assert args.module == ['des', 'galsim_extra']

    args = galsim.main.parse_args([config_file, '-p'])
    assert args.config_file == config_file
    assert args.profile is True

    args = galsim.main.parse_args([config_file, '-n', '3'])
    assert args.config_file == config_file
    assert args.njobs == 3

    args = galsim.main.parse_args([config_file, '-n', '3', '-j', '3'])
    assert args.config_file == config_file
    assert args.njobs == 3
    assert args.job == 3
    args = galsim.main.parse_args([config_file, '-n', '3', '-j', '1'])
    assert args.config_file == config_file
    assert args.njobs == 3
    assert args.job == 1

    args = galsim.main.parse_args([config_file, '-x'])
    assert args.config_file == config_file
    assert args.except_abort is True

    # --version with a config file doesn't exit, but still prints the version.
    print('This should print the version...')
    args = galsim.main.parse_args([config_file, '--version'])
    assert args.config_file == config_file

    # Additional arguments are accumulated in args.variables
    args = galsim.main.parse_args([config_file, 'output.nfiles=1', 'output.dir="."'])
    assert args.config_file == config_file
    assert args.variables == ['output.nfiles=1', 'output.dir="."']

    # Some invalid parameters
    with assert_raises(ValueError):
        galsim.main.parse_args([config_file, '-n', '0'])
    with assert_raises(ValueError):
        galsim.main.parse_args([config_file, '-n', '-1'])
    with assert_raises(ValueError):
        galsim.main.parse_args([config_file, '-n', '3', '-j', '4'])
    with assert_raises(ValueError):
        galsim.main.parse_args([config_file, '-n', '3', '-j', '0'])
    with assert_raises(ValueError):
        galsim.main.parse_args([config_file, '-j', '3'])

    # The ones handled by ArgumentParser raise SystemExit and print to stderr
    # To avoid ugly text output during pytest runs, redirect stderr to stdout for a moment.
    sys_stderr = sys.stderr
    sys.stderr = sys.stdout
    with assert_raises(SystemExit):
        galsim.main.parse_args([config_file, '-f', 'invalid'])
    with assert_raises(SystemExit):
        galsim.main.parse_args([config_file, '-v', '-1'])
    with assert_raises(SystemExit):
        galsim.main.parse_args([config_file, '-v', '4'])
    sys.stderr = sys_stderr

# Need to call these before each time make_logger is repeated.  Else duplicate handles.
def remove_handler():
    logger = logging.getLogger('galsim')
    for handler in logger.handlers:
        logger.removeHandler(handler)

@timer
def test_logger():
    args = galsim.main.parse_args(['test.yaml'])

    remove_handler()
    logger = galsim.main.make_logger(args)
    assert logger.getEffectiveLevel() == logging.WARNING
    print('handlers = ',logger.handlers)
    print('This should print...')
    logger.warning("Test warning")
    print('This should not print...')
    logger.info("Test info")
    logger.debug("Test debug")
    print('Done')

    args.verbosity = 3
    args.log_file = 'output/test_logger.log'
    remove_handler()
    logger = galsim.main.make_logger(args)
    print('handlers = ',logger.handlers)
    assert logger.getEffectiveLevel() == logging.DEBUG
    logger.warning("Test warning")
    logger.info("Test info")
    logger.debug("Test debug")

    with open(args.log_file) as f:
        assert f.readline().strip() == "Test warning"
        assert f.readline().strip() == "Test info"
        assert f.readline().strip() == "Test debug"

    args.verbosity = 0
    remove_handler()
    logger = galsim.main.make_logger(args)
    print('handlers = ',logger.handlers)
    assert logger.getEffectiveLevel() == logging.ERROR
    logger.warning("Test warning")
    logger.info("Test info")
    logger.debug("Test debug")

    with open(args.log_file) as f:
        assert f.readlines() == []

    args.verbosity = 3
    args.log_format = '%(levelname)s - %(message)s'
    remove_handler()
    logger = galsim.main.make_logger(args)
    print('handlers = ',logger.handlers)
    assert logger.getEffectiveLevel() == logging.DEBUG
    logger.warning("Test warning")
    logger.info("Test info")
    logger.debug("Test debug")

    with open(args.log_file) as f:
        assert f.readline().strip() == "WARNING - Test warning"
        assert f.readline().strip() == "INFO - Test info"
        assert f.readline().strip() == "DEBUG - Test debug"

@timer
def test_parse_variables():
    logger = logging.getLogger('test_main')
    logger.setLevel(logging.ERROR)

    # Empty list -> empty dict
    new_params = galsim.main.parse_variables([], logger)
    assert new_params == {}

    vars = ["output.nfiles=1", "output.dir='.'"]
    new_params = galsim.main.parse_variables(vars, logger)
    assert new_params['output.nfiles'] == 1
    assert new_params['output.dir'] == '.'

    # Lists or dicts will be parsed here
    vars = ["psf={'type':'Gaussian','sigma':0.4}", "output.skip=[0,0,0,0,0,1]"]
    new_params = galsim.main.parse_variables(vars, logger)
    assert new_params['psf'] == {'type' : 'Gaussian', 'sigma' : 0.4}
    assert new_params['output.skip'] == [0,0,0,0,0,1]

    # Things that don't parse properly are just returned verbatim as a string
    # (Presumably they would give an appropriate error later when they are used if the
    # string is not a valid value for whatever this is.)
    vars = ["psf={'type':'Gaussian' : 'sigma':0.4}"]
    new_params = galsim.main.parse_variables(vars, logger)
    assert new_params['psf'] == "{'type':'Gaussian' : 'sigma':0.4}"

    # Missing = is an error
    vars = ["output.nfiles","1"]
    assert_raises(galsim.GalSimError, galsim.main.parse_variables, vars, logger)
    vars = ["output.nfiles-1"]
    assert_raises(galsim.GalSimError, galsim.main.parse_variables, vars, logger)

    # Should work correctly if yaml isn't available.
    # Although it doesn't always parse quite as nicely. E.g. requires ", not ' for string quotes.
    with mock.patch.dict(sys.modules, {'yaml':None}):
        vars = ['psf={"type":"Gaussian","sigma":0.4}', 'output.skip=[0,0,0,0,0,1]']
        new_params = galsim.main.parse_variables(vars, logger)
        assert new_params['psf'] == {'type' : 'Gaussian', 'sigma' : 0.4}
        assert new_params['output.skip'] == [0,0,0,0,0,1]

@timer
def test_modules():
    config = {}
    galsim.main.add_modules(config, None)
    assert config == {}

    galsim.main.add_modules(config, [])
    assert config == {}

    galsim.main.add_modules(config, ['des'])
    assert config == {'modules' : ['des']}

    galsim.main.add_modules(config, ['galsim_extra'])
    assert config == {'modules' : ['des', 'galsim_extra']}

    galsim.main.add_modules(config, ['a','b','c'])
    assert config == {'modules' : ['des', 'galsim_extra', 'a', 'b', 'c']}

@timer
def test_process():
    # Test running an extremely simple config, just testing a few features of how
    # the process_config() function works.
    config = {
        'gal' : {
            'type' : 'Gaussian',
            'sigma' : 2,
        },
        'output' : {
            'dir' : 'output',
            'file_name' : 'test_main.fits',
        }
    }
    file_name = os.path.join('output', 'test_main.fits')
    if os.path.exists(file_name):
        os.remove(file_name)
    args = galsim.main.parse_args(['test.yaml','-v','1'])
    remove_handler()
    logger = galsim.main.make_logger(args)

    galsim.main.process_config([config], args, logger)
    assert os.path.exists(file_name)
    assert config['root'] == 'test'  # This is set automatically
    args.profile = True
    print('Should print profile:')
    galsim.main.process_config([config], args, logger)
    assert config['profile'] is True
    print('Done')

    remove_handler()
    os.remove(file_name)
    config_file = os.path.join('input','test.yaml')
    galsim.main.main([config_file,'-v','1'])
    assert os.path.exists(file_name)

    with mock.patch('sys.argv', ['galsim', config_file, '-v', '1']):
        remove_handler()
        os.remove(file_name)
        galsim.main.run_main()
        assert os.path.exists(file_name)

        remove_handler()
        os.remove(file_name)
        galsim.__main__.run_main()
        assert os.path.exists(file_name)

if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from ._galsim import j0, j1, jv, kv, yv, iv, j0_root, jv_root

# Alias the "n" names, which don't get any advantage from being implemented differently,
# so we only have the generic nu implementation.  But to match scipy.special, we also
# allow the user to write jn, kn, or yn.  No in, since that's a reserved word, of course.
jn = jv
kn = kv
yn = yv

# These aren't Bessel related, but they are similarly useful math functions that we have
# implemented in C++.  Exposing them primarily helps for testing that they give essentially
# equivalent answers as scipy.special.
from ._galsim import gammainc, sinc, si, ci



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'RealGalaxy', 'RealGalaxyCatalog', 'ChromaticRealGalaxy', ]

import os
import numpy as np
import copy
from multiprocessing import Lock

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .bounds import BoundsI
from .utilities import lazy_property, doc_inherit, convert_interpolant, merge_sorted
from .interpolant import Quintic
from .interpolatedimage import InterpolatedImage, _InterpolatedKImage
from .convolve import Convolve, Deconvolve
from .image import Image, ImageCD
from .correlatednoise import UncorrelatedNoise, BaseCorrelatedNoise, CovarianceSpectrum
from .errors import GalSimValueError, GalSimIncompatibleValuesError
from .errors import GalSimIndexError
from .table import LookupTable
from .random import BaseDeviate, UniformDeviate
from .sed import SED
from .bandpass import Bandpass
from .chromatic import ChromaticSum
from . import meta_data
from ._pyfits import pyfits


HST_area = 45238.93416  # Area of HST primary mirror in cm^2 from Synphot User's Guide.

# Currently, have bandpasses available for HST COSMOS, AEGIS, and CANDELS.
# ACS zeropoints (AB magnitudes) from
# http://www.stsci.edu/hst/acs/analysis/zeropoints/old_page/localZeropoints#tablestart
# WFC3 zeropoints (AB magnitudes) from
# http://www.stsci.edu/hst/wfc3/phot_zp_lbn
# Format of dictionary entry is:
#    'KEY' : tuple(bandpass filename, zeropoint)
real_galaxy_bandpasses = {
        'F275W': ('WFC3_uvis_F275W.dat', 24.1305),
        'F336W': ('WFC3_uvis_F336W.dat', 24.6682),
        'F435W': ('ACS_wfc_F435W.dat', 25.65777),
        'F606W': ('ACS_wfc_F606W.dat', 26.49113),
        'F775W': ('ACS_wfc_F775W.dat', 25.66504),
        'F814W': ('ACS_wfc_F814W.dat', 25.94333),
        'F850LP': ('ACS_wfc_F850LP.dat', 24.84245),
        'F105W': ('WFC3_ir_F105W.dat', 26.2687),
        'F125W': ('WFC3_ir_F125W.dat', 26.2303),
        'F160W': ('WFC3_ir_F160W.dat', 25.9463)
}

class RealGalaxy(GSObject):
    """A class describing real galaxies from some training dataset.  Its underlying implementation
    uses a Convolution instance of an `InterpolatedImage` (for the observed galaxy) with a
    `Deconvolution` of another `InterpolatedImage` (for the PSF).

    This class uses a catalog describing galaxies in some training data (for more details, see the
    `RealGalaxyCatalog` documentation) to read in data about realistic galaxies that can be used for
    simulations based on those galaxies.  Also included in the class is additional information that
    might be needed to make or interpret the simulations, e.g., the noise properties of the training
    data.  Users who wish to draw RealGalaxies that have well-defined flux scalings in various
    passbands, and/or parametric representations, should use the COSMOSGalaxy class.

    Because RealGalaxy involves a `Deconvolution`, ``method = 'phot'`` is unavailable for the
    `GSObject.drawImage` function, and it is essential that users convolve each RealGalaxy with a
    PSF that is at least as large as the original HST PSF (stored as an attribute) before rendering
    any images.  This is necessary to eliminate noise that was amplified due to deconvolution of the
    HST PSF.

    Example::

        >>> real_galaxy = galsim.RealGalaxy(real_galaxy_catalog, index=None, id=None, random=False,
        ...                                 rng=None, x_interpolant=None, k_interpolant=None,
        ...                                 flux=None, pad_factor=4, noise_pad_size=0,
        ...                                 gsparams=None)

    This initializes ``real_galaxy`` with three `InterpolatedImage` objects (one for the deconvolved
    galaxy, and saved versions of the original HST image and PSF). Note that there are multiple
    keywords for choosing a galaxy; exactly one must be set.

    Note that tests suggest that for optimal balance between accuracy and speed, ``k_interpolant``
    and ``pad_factor`` should be kept at their default values.  The user should be aware that
    significant inaccuracy can result from using other combinations of these parameters; more
    details can be found in http://arxiv.org/abs/1401.2636, especially table 1, and in comment
    https://github.com/GalSim-developers/GalSim/issues/389#issuecomment-26166621 and the following
    comments.

    If you don't set a flux, the flux of the returned object will be the flux of the original
    HST data, scaled to correspond to a 1 second HST exposure (though see the ``area_norm``
    parameter below, and also caveats related to using the ``flux`` parameter).  If you want a flux
    appropriate for a longer exposure, or for a telescope with a different collecting area than HST,
    you can either renormalize the object with the ``flux_rescale`` parameter, or by using the
    ``exptime`` and ``area`` parameters to `GSObject.drawImage`.

    Note that RealGalaxy objects use arcsec for the units of their linear dimension.  If you
    are using a different unit for other things (the PSF, WCS, etc.), then you should dilate
    the resulting object with ``gal.dilate(galsim.arcsec / scale_unit)``.

    Parameters:
        real_galaxy_catalog:    A `RealGalaxyCatalog` object with basic information about where to
                                find the data, etc.
        index:                  Index of the desired galaxy in the catalog. [One of ``index``,
                                ``id``, or ``random`` is required.]
        id:                     Object ID for the desired galaxy in the catalog. [One of ``index``,
                                ``id``, or ``random`` is required.]
        random:                 If True, then select a random galaxy from the catalog.  If the
                                catalog has a 'weight' associated with it to allow for correction of
                                selection effects in which galaxies were included, the 'weight'
                                factor is used to remove those selection effects rather than
                                selecting a completely random object.
                                [One of ``index``, ``id``, or ``random`` is required.]
        rng:                    A random number generator to use for selecting a random galaxy
                                (may be any kind of `BaseDeviate` or None) and to use in generating
                                any noise field when padding. [default: None]
        x_interpolant:          Either an `Interpolant` instance or a string indicating which
                                real-space interpolant should be used.  Options are 'nearest',
                                'sinc', 'linear', 'cubic', 'quintic', or 'lanczosN' where N should
                                be the integer order to use. [default: galsim.Quintic()]
        k_interpolant:          Either an `Interpolant` instance or a string indicating which
                                k-space interpolant should be used.  Options are 'nearest', 'sinc',
                                'linear', 'cubic', 'quintic', or 'lanczosN' where N should be the
                                integer order to use.  We strongly recommend leaving this parameter
                                at its default value; see text above for details.
                                [default: galsim.Quintic()]
        flux:                   Total flux, if None then original flux in image is adopted without
                                change.  Note that, technically, this parameter sets the flux of the
                                postage stamp image and not the flux of the contained galaxy.
                                These two values will be strongly correlated when the signal-to-
                                noise ratio of the galaxy is large, but may be considerably
                                different if the flux of the galaxy is small with respect to the
                                noise variations in the postage stamp.  To avoid complications with
                                faint galaxies, consider using the flux_rescale parameter.
                                [default: None]
        flux_rescale:           Flux rescaling factor; if None, then no rescaling is done.  Either
                                ``flux`` or ``flux_rescale`` may be set, but not both.
                                [default: None]
        pad_factor:             Factor by which to pad the `Image` when creating the
                                `InterpolatedImage`.  We strongly recommend leaving this parameter
                                at its default value; see text above for details.  [default: 4]
        noise_pad_size:         If provided, the image will be padded out to this size (in arcsec)
                                with the noise specified in the real galaxy catalog. This is
                                important if you are planning to whiten the resulting image.  You
                                should make sure that the padded image is larger than the postage
                                stamp onto which you are drawing this object.
                                [default: None]
        area_norm:              Area in cm^2 by which to normalize the flux of the returned object.
                                When area_norm=1 (the default), drawing with `GSObject.drawImage`
                                keywords exptime=1 and area=1 will simulate an image with the
                                appropriate number of counts for a 1 second exposure with the
                                original telescope/camera (e.g., with HST when using the COSMOS
                                catalog).
                                If you would rather explicitly specify the collecting area of the
                                telescope when using `GSObject.drawImage` with a `RealGalaxy`,
                                then you should set area_norm equal to the collecting area of the
                                source catalog telescope when creating the `RealGalaxy` (e.g.,
                                area_norm=45238.93416 for HST).  [default: 1]
        gsparams:               An optional `GSParams` argument. [default: None]
        logger:                 A logger object for output of progress statements if the user wants
                                them.  [default: None]
    """
    _opt_params = { "x_interpolant" : str ,
                    "k_interpolant" : str ,
                    "flux" : float ,
                    "flux_rescale" : float ,
                    "pad_factor" : float,
                    "noise_pad_size" : float,
                    "area_norm" : float
                  }
    _single_params = [ { "index" : int , "id" : str , "random" : bool } ]
    _takes_rng = True

    _has_hard_edges = False
    _is_axisymmetric = False
    _is_analytic_x = False
    _is_analytic_k = True

    def __init__(self, real_galaxy_catalog, index=None, id=None, random=False,
                 rng=None, x_interpolant=None, k_interpolant=None, flux=None, flux_rescale=None,
                 pad_factor=4, noise_pad_size=0, area_norm=1.0, gsparams=None, logger=None):
        if rng is None:
            rng = BaseDeviate()
        elif not isinstance(rng, BaseDeviate):
            raise TypeError("The rng provided to RealGalaxy is not a BaseDeviate")
        self.rng = rng

        if flux is not None and flux_rescale is not None:
            raise GalSimIncompatibleValuesError(
                "Cannot supply a flux and a flux rescaling factor.",
                flux=flux, flux_rescale=flux_rescale)

        logger = LoggerWrapper(logger)  # So don't need to check `if logger:` all the time.

        if isinstance(real_galaxy_catalog, tuple):
            # Special (undocumented) way to build a RealGalaxy without needing the rgc directly
            # by providing the things we need from it.  Used by COSMOSGalaxy.
            self.gal_image, self.psf_image, noise_image, pixel_scale, var = real_galaxy_catalog
            use_index = 0  # For the logger statements below.
            logger.debug('RealGalaxy %d: Start RealGalaxy constructor.',use_index)
            self.catalog_file = None
            self.catalog = ''
        else:
            # Get the index to use in the catalog
            if index is not None:
                if id is not None or random:
                    raise GalSimIncompatibleValuesError(
                        "Too many methods for selecting a galaxy.",
                        index=index, id=id, random=random)
                use_index = index
            elif id is not None:
                if random:
                    raise GalSimIncompatibleValuesError(
                        "Too many methods for selecting a galaxy.", id=id, random=random)
                use_index = real_galaxy_catalog.getIndexForID(id)
            elif random:
                ud = UniformDeviate(self.rng)
                use_index = int(real_galaxy_catalog.nobjects * ud())
                if real_galaxy_catalog.weight is not None:
                    # If weight factors are available, make sure the random selection uses the
                    # weights to remove the catalog-level selection effects (flux_radius-dependent
                    # probability of making a postage stamp for a given object).
                    while ud() > real_galaxy_catalog.weight[use_index]:
                        # Pick another one to try.
                        use_index = int(real_galaxy_catalog.nobjects * ud())
            else:
                raise GalSimIncompatibleValuesError(
                    "No method specified for selecting a galaxy.",
                    index=index, id=id, random=random)
            logger.debug('RealGalaxy %d: Start RealGalaxy constructor.',use_index)

            # Read in the galaxy, PSF images; for now, rely on pyfits to make I/O errors.
            self.gal_image = real_galaxy_catalog.getGalImage(use_index)
            logger.debug('RealGalaxy %d: Got gal_image',use_index)

            self.psf_image = real_galaxy_catalog.getPSFImage(use_index)
            logger.debug('RealGalaxy %d: Got psf_image',use_index)

            #self._gal_noise = real_galaxy_catalog.getNoise(use_index, self.rng, gsparams)
            # We need to duplication some of the RealGalaxyCatalog.getNoise() function, since we
            # want it to be possible to have the RealGalaxyCatalog in another process, and the
            # BaseCorrelatedNoise object is not picklable.  So we just build it here instead.
            noise_image, pixel_scale, var = real_galaxy_catalog.getNoiseProperties(use_index)
            logger.debug('RealGalaxy %d: Got noise_image',use_index)
            self.catalog_file = real_galaxy_catalog.getFileName()
            self.catalog = real_galaxy_catalog

        self._gsparams = GSParams.check(gsparams)

        if noise_image is None:
            self._gal_noise = UncorrelatedNoise(var, rng=self.rng, scale=pixel_scale,
                                                gsparams=self._gsparams)
        else:
            ii = InterpolatedImage(noise_image, normalization="sb",
                                   calculate_stepk=False, calculate_maxk=False,
                                   x_interpolant='linear', gsparams=self._gsparams)
            self._gal_noise = BaseCorrelatedNoise(self.rng, ii, noise_image.wcs)
            self._gal_noise = self._gal_noise.withVariance(var)
        logger.debug('RealGalaxy %d: Finished building noise',use_index)

        # Save any other relevant information as instance attributes
        self.index = use_index
        self.pixel_scale = float(pixel_scale)
        self._x_interpolant = x_interpolant
        self._k_interpolant = k_interpolant
        self._pad_factor = pad_factor
        self._noise_pad_size = noise_pad_size
        self._input_flux = flux
        self._flux_rescale = flux_rescale
        self._area_norm = area_norm

        # Convert noise_pad to the right noise to pass to InterpolatedImage
        if noise_pad_size:
            noise_pad = self._gal_noise
        else:
            noise_pad = 0.

        # Build the InterpolatedImage of the PSF.
        self.original_psf = InterpolatedImage(
            self.psf_image, x_interpolant=x_interpolant, k_interpolant=k_interpolant,
            flux=1.0, gsparams=self._gsparams)
        logger.debug('RealGalaxy %d: Made original_psf',use_index)

        # Build the InterpolatedImage of the galaxy.
        # Use the stepk value of the PSF as a maximum value for stepk of the galaxy.
        # (Otherwise, low surface brightness galaxies can get a spuriously high stepk, which
        # leads to problems.)
        self.original_gal = InterpolatedImage(
                self.gal_image, x_interpolant=x_interpolant, k_interpolant=k_interpolant,
                pad_factor=pad_factor, noise_pad_size=noise_pad_size,
                calculate_stepk=self.original_psf.stepk,
                calculate_maxk=self.original_psf.maxk,
                noise_pad=noise_pad, rng=self.rng, gsparams=self._gsparams)
        logger.debug('RealGalaxy %d: Made original_gal',use_index)

        # Only alter normalization if a change is requested
        if flux is not None or flux_rescale is not None or area_norm != 1:
            if flux_rescale is None:
                flux_rescale = 1.0
            flux_rescale /= area_norm
            if flux is not None:
                flux_rescale *= flux/self.original_gal.flux
            self.original_gal *= flux_rescale
            self._gal_noise *= flux_rescale**2

        logger.debug('RealGalaxy %d: Finished building RealGalaxy',use_index)

    @doc_inherit
    def withGSParams(self, gsparams=None, **kwargs):
        if gsparams == self.gsparams: return self
        ret = copy.copy(self)
        ret._gsparams = GSParams.check(gsparams, **kwargs)
        ret.original_gal = self.original_gal.withGSParams(ret._gsparams, **kwargs)
        ret.original_psf = self.original_psf.withGSParams(ret._gsparams, **kwargs)
        ret._gal_noise = self._gal_noise.withGSParams(ret._gsparams, **kwargs)
        return ret

    @classmethod
    def makeFromImage(cls, image, PSF, xi, **kwargs):
        """Create a `RealGalaxy` directly from image, PSF, and noise description.

        Parameters:
            image:  `Image` of the galaxy you want to simulate.
            PSF:    `GSObject` representing the PSF of the galaxy image.  Note that this PSF
                    should include the response of the pixel convolution.
            xi:     `BaseCorrelatedNoise` object characterizing the noise correlations in the input
                    image.
        """
        noise_image = xi.drawImage()
        pixel_scale = noise_image.scale
        var = xi.getVariance()
        psf_image = PSF.drawImage(method='no_pixel')
        return RealGalaxy((image, psf_image, noise_image, pixel_scale, var))

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, RealGalaxy) and
                 self.catalog == other.catalog and
                 self.index == other.index and
                 self._x_interpolant == other._x_interpolant and
                 self._k_interpolant == other._k_interpolant and
                 self._pad_factor == other._pad_factor and
                 self._noise_pad_size == other._noise_pad_size and
                 self._input_flux == other._input_flux and
                 self._flux_rescale == other._flux_rescale and
                 self._area_norm == other._area_norm and
                 self._gsparams == other._gsparams))

    def __hash__(self):
        return hash(("galsim.RealGalaxy", self.catalog, self.index, self._x_interpolant,
                     self._k_interpolant, self._pad_factor, self._noise_pad_size, self._input_flux,
                     self._flux_rescale, self._area_norm, self._gsparams))

    def __repr__(self):
        s = 'galsim.RealGalaxy(%r, index=%r, '%(self.catalog, self.index)
        if self._x_interpolant is not None:
            s += 'x_interpolant=%r, '%self._x_interpolant
        if self._k_interpolant is not None:
            s += 'k_interpolant=%r, '%self._k_interpolant
        if self._pad_factor != 4:
            s += 'pad_factor=%r, '%self._pad_factor
        if self._noise_pad_size != 0:
            s += 'noise_pad_size=%r, '%self._noise_pad_size
        if self._input_flux is not None:
            s += 'flux=%r, '%self._input_flux
        if self._flux_rescale is not None:
            s += 'flux_rescale=%r, '%self._flux_rescale
        if self._area_norm != 1:
            s += 'area_norm=%r, '%self._area_norm
        s += 'rng=%r, '%self.rng
        s += 'gsparams=%r)'%self._gsparams
        return s

    def __str__(self):
        # I think this is more intuitive without the RealGalaxyCatalog parameter listed.
        return 'galsim.RealGalaxy(index=%s, flux=%s)'%(self.index, self.flux)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_conv',None)
        d.pop('_psf_inv',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @lazy_property
    def _psf_inv(self):
        return Deconvolve(self.original_psf, gsparams=self._gsparams)

    @lazy_property
    def _conv(self):
        return Convolve([self.original_gal, self._psf_inv], gsparams=self._gsparams)

    @property
    def _noise(self):
        # We just store the original noise, not convolved with psf_inv until we need it,
        # mostly so we don't have to invalidate this if gsparams changes.
        return self._gal_noise.convolvedWith(self._psf_inv, self._gsparams)

    @property
    def _maxk(self):
        return self._conv._maxk

    @property
    def _stepk(self):
        return self._conv._stepk

    @property
    def _centroid(self):
        return self._conv._centroid

    @property
    def _flux(self):
        return self._conv._flux

    @property
    def _positive_flux(self):
        return self._conv._positive_flux

    @property
    def _negative_flux(self):
        return self._conv._negative_flux

    @lazy_property
    def _flux_per_photon(self):
        return self._calculate_flux_per_photon()

    @property
    def _max_sb(self):
        return self._conv._max_sb

    def _kValue(self, kpos):
        return self._conv._kValue(kpos)

    def _drawKImage(self, image, jac=None):
        self._conv._drawKImage(image, jac)


class RealGalaxyCatalog:
    """Class containing a catalog with information about real galaxy training data.

    The RealGalaxyCatalog class reads in and stores information about a specific training sample of
    realistic galaxies. We assume that all files containing the images (galaxies and PSFs) live in
    one directory; they could be individual files, or multiple HDUs of the same file.  Currently
    there is no functionality that lets this be a FITS data cube, because we assume that the object
    postage stamps will in general need to be different sizes depending on the galaxy size.

    Note that when simulating galaxies based on HST but using either realistic or parametric galaxy
    models, the COSMOSCatalog class may be more useful.  It allows the imposition of selection
    criteria and other subtleties that are more difficult to impose with RealGalaxyCatalog.

    While you could create your own catalog to use with this class, the typical use cases would
    be to use one of the catalogs that we have created and distributed.  There are three such
    catalogs currently, which can be use with one of the following initializations:

    1. A small example catalog is distributed with the GalSim distribution.  This catalog only
       has 100 galaxies, so it is not terribly useful as a representative galaxy population.
       But for simplistic use cases, it might be sufficient.  We use it for our unit tests and
       in some of the demo scripts (demo6, demo10, and demo11).  To use this catalog, you would
       initialize with::

           >>> rgc = galsim.RealGalaxyCatalog('real_galaxy_catalog_23.5_example.fits',
                                              dir='path/to/GalSim/examples/data')

    2. There are two larger catalogs based on HST observations of the COSMOS field with around
       26,000 and 56,000 galaxies each with a limiting magnitude of F814W=23.5.  (The former is
       a subset of the latter.) For information about how to download these catalogs, see the
       RealGalaxy Data Download Page on the GalSim Wiki:

       https://github.com/GalSim-developers/GalSim/wiki/RealGalaxy%20Data

       Be warned that the catalogs are quite large.  The larger one is around 11 GB after unpacking
       the tarball.  To use one of these catalogs, you would initialize with::

           >>> rgc = galsim.RealGalaxyCatalog('real_galaxy_catalog_23.5.fits',
                                              dir='path/to/download/directory')

    3. There is a catalog containing a random subsample of the HST COSMOS images with a limiting
       magnitude of F814W=25.2.  More information about downloading these catalogs can be found on
       the RealGalaxy Data Download page linked above.

    4. Finally, we provide a program that will download the large COSMOS sample for you and
       put it in the $PREFIX/share/galsim directory of your installation path.  The program is::

           galsim_download_cosmos

       which gets installed in the $PREFIX/bin directory when you install GalSim.  If you use
       this program to download the COSMOS catalog, then you can use it with::

           >>> rgc = galsim.RealGalaxyCatalog()

       GalSim knows the location of the installation share directory, so it will automatically
       look for it there.

    .. note::

        A RealGalaxyCatalog instance will cache data from the input files that it loads.
        Normally, this leads to a performance improvement, especially if objects are selected
        from the catalog semi-randomly, since it is faster to access the data in memory.
        If you want to free the cached memory before the RealGalaxyCatalog goes out of
        scope, you can use the `close` method to do so.

    Parameters:
        file_name:  The file containing the catalog. [default: None, which will look for the
                    F814W<25.2 COSMOS catalog in $PREFIX/share/galsim.  It will raise an
                    exception if the catalog is not there telling you to run
                    galsim_download_cosmos.]
        sample:     A keyword argument that can be used to specify the sample to use, i.e.,
                    "23.5" or "25.2".  At most one of ``file_name`` and ``sample`` should be
                    specified.
                    [default: None, which results in the same default as ``file_name=None``.]
        dir:        The directory containing the catalog, image, and noise files, or symlinks to
                    them. [default: None]
        preload:    Whether to preload the header information.  If ``preload=True``, the bulk of
                    the I/O time is in the constructor.  If ``preload=False``, there is
                    approximately the same total I/O time (assuming you eventually use most of
                    the image files referenced in the catalog), but it is spread over the
                    various calls to `getGalImage` and `getPSFImage`.  [default: False]
        logger:     An optional logger object to log progress. [default: None]
    """
    _opt_params = { 'file_name' : str, 'sample' : str, 'dir' : str,
                    'preload' : bool }

    # _nobject_only is an intentionally undocumented kwarg that should be used only by
    # the config structure.  It indicates that all we care about is the nobjects parameter.
    # So skip any other calculations that might normally be necessary on construction.
    def __init__(self, file_name=None, sample=None, dir=None, preload=False, logger=None):
        if sample is not None and file_name is not None:
            raise GalSimIncompatibleValuesError(
                "Cannot specify both the sample and file_name.",
                sample=sample, file_name=file_name)

        logger = LoggerWrapper(logger)

        self.file_name, self.image_dir, self.sample = _parse_files_dirs(file_name, dir, sample)

        with pyfits.open(self.file_name) as fits:
            self.cat = fits[1].data
        self.nobjects = len(self.cat) # number of objects in the catalog
        logger.debug('RealGalaxyCatalog %s has %d objects',self.file_name,self.nobjects)

        self._preload = preload
        self.saved_noise_im = {}
        self.loaded_files = {}
        # The pyfits commands aren't thread safe.  So we need to make sure the methods that
        # use pyfits are not run concurrently from multiple threads.
        self.gal_lock = Lock()  # Use this when accessing gal files
        self.psf_lock = Lock()  # Use this when accessing psf files
        self.loaded_lock = Lock()  # Use this when opening new files from disk
        self.noise_lock = Lock()  # Use this for building the noise image(s) (usually just one)


    # Some lazy properties that we set up the first time they are used.
    @lazy_property
    def ident(self):
        ident = self.cat.field('ident') # ID for object in the training sample
        # We want to make sure that the ident array contains all strings.
        # Strangely, ident.astype(str) produces a string with each element == '1'.
        # Hence this way of doing the conversion:
        return [ "%s"%val for val in ident ]

    @lazy_property
    def gal_file_name(self):
        gal_file_name = self.cat.field('gal_filename') # file containing the galaxy image
        # Add the directories:
        # Note the strip call.  Sometimes the filenames have an extra space at the end.
        # This gets rid of that space.
        return [os.path.join(self.image_dir,f.strip()) for f in gal_file_name]

    @lazy_property
    def psf_file_name(self):
        psf_file_name = self.cat.field('PSF_filename') # file containing the PSF image
        return [os.path.join(self.image_dir,f.strip()) for f in psf_file_name]

    @lazy_property
    def noise_file_name(self):
        # We don't require the noise_filename column.  If it is not present, we will use
        # Uncorrelated noise based on the variance column.
        try:
            noise_file_name = self.cat.field('noise_filename') # file containing the noise cf
        except KeyError:
            return None
        else:
            return [os.path.join(self.image_dir,f) for f in noise_file_name]

    @lazy_property
    def gal_hdu(self):
        return self.cat.field('gal_hdu') # HDU containing the galaxy image

    @lazy_property
    def psf_hdu(self):
        return self.cat.field('PSF_hdu') # HDU containing the PSF image

    @lazy_property
    def pixel_scale(self):
        return self.cat.field('pixel_scale') # pixel scale for image (could be different
        # if we have training data from other datasets... let's be general here and make it a
        # vector in case of mixed training set)

    @lazy_property
    def variance(self):
        return self.cat.field('noise_variance') # noise variance for image

    @lazy_property
    def mag(self):
        return self.cat.field('mag')   # apparent magnitude

    @lazy_property
    def band(self):
        return self.cat.field('band') # bandpass in which apparent mag is measured, e.g., F814W

    @lazy_property
    def weight(self):
        # The weight factor should be a float value >=0 (so that random selections of indices can
        # use it to remove any selection effects in the catalog creation process).
        # Here we renormalize by the maximum weight.  If the maximum is below 1, that just means
        # that all galaxies were subsampled at some level, and here we only want to account for
        # relative selection effects within the catalog, not absolute subsampling.  If the maximum
        # is above 1, then our random number generation test used to draw a weighted sample will
        # fail since we use uniform deviates in the range 0 to 1.
        try:
            weight = self.cat.field('weight')
        except KeyError:  # pragma: no cover
            raise OSError("You still have the old COSMOS catalog.  Run the program "
                          "`galsim_download_cosmos -s %s` to upgrade."%(self.sample))
        else:
            return weight/np.max(weight)

    @lazy_property
    def stamp_flux(self):
        try:
            return self.cat.field('stamp_flux')
        except KeyError:  # pragma: no cover
            raise OSError("You still have the old COSMOS catalog.  Run the program "
                          "`galsim_download_cosmos -s %s` to upgrade."%(self.sample))

    def __del__(self):
        # Make sure to clean up pyfits open files if people forget to call close()
        self.close()

    def close(self):
        """Close all loaded files.

        Usually it is a performance improvement to keep the various input files open while using
        the RealGalaxyCatalog, especially if the access of the galaxies is in semi-random order.
        However, if you want to release the memory tied up in these files before the
        RealGalaxyCatalog goes out of scope, then this method will release the memory directly.
        """
        # Need to close any open files.
        # Make sure to check if loaded_files exists, since the constructor could abort
        # before it gets to the place where loaded_files is built.
        if hasattr(self, 'loaded_files'):
            for f in self.loaded_files.values():
                f.close()
            self.loaded_files.clear()
            # Also release the memory in the noise images, even though this is usually negligible.
            self.saved_noise_im.clear()

    def getNObjects(self) : return self.nobjects
    def __len__(self): return self.nobjects
    def getFileName(self) : return self.file_name

    def getIndexForID(self, id):
        """Internal function to find which index number corresponds to the value ID in the ident
        field.
        """
        # Just to be completely consistent, convert id to a string in the same way we
        # did above for the ident array:
        id = "%s"%id
        if id in self.ident:
            return self.ident.index(id)
        else:
            raise GalSimValueError('ID not found in list of IDs',id, self.ident)

    def _maybe_preload(self):
        # Preload all files if desired.
        # This is delayed until the first time we might need it, since we might only need
        # to know nobjects and not load the data at all.  The first time we try to do something
        # that needs the files, we'll call preload (if requested).
        if self._preload:
            self.preload()
            self._preload = False  # Once we've loaded them.  Don't do it again.

    def preload(self):
        """Preload the files into memory.

        There are memory implications to this, so we don't do this by default.  However, it can be
        a big speedup if memory isn't an issue.
        """
        with self.loaded_lock:
            for file_name in np.concatenate((self.gal_file_name , self.psf_file_name)):
                # numpy sometimes add a space at the end of the string that is not present in
                # the original file.  Stupid.  But this next line removes it.
                file_name = file_name.strip()
                if file_name not in self.loaded_files:
                    # I use memmap=False, because I was getting problems with running out of
                    # file handles in the great3 real_gal run, which uses a lot of rgc files.
                    # I think there must be a bug in pyfits that leaves file handles open somewhere
                    # when memmap = True.  Anyway, I don't know what the performance implications
                    # are (since I couldn't finish the run with the default memmap=True), but I
                    # don't think there is much impact either way with memory mapping in our case.
                    f = pyfits.open(file_name,memmap=False)
                    self.loaded_files[file_name] = f
                    # Access all the data from all hdus to force PyFits to read the data
                    for hdu in f:
                        hdu.data

    def _getFile(self, file_name):
        self._maybe_preload()
        if file_name in self.loaded_files:
            f = self.loaded_files[file_name]
        else:
            with self.loaded_lock:
                # Check again in case two processes both hit the else at the same time.
                if file_name in self.loaded_files: # pragma: no cover
                    f = self.loaded_files[file_name]
                else:
                    f = pyfits.open(file_name,memmap=False)
                    self.loaded_files[file_name] = f
        return f

    def getBandpass(self):
        """Returns a `Bandpass` object for the catalog.
        """
        try:
            bp = real_galaxy_bandpasses[self.band[0].upper()]
        except KeyError:
            raise GalSimValueError("Bandpass not found.  To use this bandpass, please add an entry "
                                   "to the galsim.real.real_galaxy_bandpasses dictionary.",
                                   self.band[0], real_galaxy_bandpasses.keys())
        return Bandpass(bp[0], wave_type='nm', zeropoint=bp[1])

    def getGalImage(self, i):
        """Returns the galaxy at index ``i`` as an `Image` object.
        """
        if i >= len(self.gal_file_name):
            raise GalSimIndexError('index out of range (0..%d)'%(len(self.gal_file_name)-1),i)
        f = self._getFile(self.gal_file_name[i])
        with self.gal_lock:
            array = f[self.gal_hdu[i]].data
        im = Image(np.ascontiguousarray(array.astype(np.float64)), scale=self.pixel_scale[i])
        return im

    def getPSFImage(self, i):
        """Returns the PSF at index ``i`` as an `Image` object.
        """
        if i >= len(self.psf_file_name):
            raise GalSimIndexError('index out of range (0..%d)'%(len(self.psf_file_name)-1),i)
        f = self._getFile(self.psf_file_name[i])
        with self.psf_lock:
            array = f[self.psf_hdu[i]].data
        return Image(np.ascontiguousarray(array.astype(np.float64)), scale=self.pixel_scale[i])

    def getPSF(self, i, x_interpolant=None, k_interpolant=None, gsparams=None):
        """Returns the PSF at index ``i`` as a `GSObject`.
        """
        psf_image = self.getPSFImage(i)
        return InterpolatedImage(psf_image,
                                 x_interpolant=x_interpolant, k_interpolant=k_interpolant,
                                 flux=1.0, gsparams=gsparams)

    def getNoiseProperties(self, i):
        """Returns the components needed to make the noise correlation function at index ``i``.
           Specifically, the noise image (or None), the pixel_scale, and the noise variance,
           as a tuple (im, scale, var).
        """
        if self.noise_file_name is None:
            im = None
        else:
            if i >= len(self.noise_file_name):
                raise GalSimIndexError('index out of range (0..%d)'%(len(self.noise_file_name)-1),i)
            if self.noise_file_name[i] in self.saved_noise_im:
                im = self.saved_noise_im[self.noise_file_name[i]]
            else:
                with self.noise_lock:
                    # Again, a second check in case two processes get here at the same time.
                    if self.noise_file_name[i] in self.saved_noise_im:  # pragma: no cover
                        im = self.saved_noise_im[self.noise_file_name[i]]
                    else:
                        with pyfits.open(self.noise_file_name[i]) as fits:
                            array = fits[0].data
                        im = Image(np.ascontiguousarray(array.astype(np.float64)),
                                        scale=self.pixel_scale[i])
                        self.saved_noise_im[self.noise_file_name[i]] = im

        return im, self.pixel_scale[i], self.variance[i]

    def getNoise(self, i, rng=None, gsparams=None):
        """Returns the noise correlation function at index ``i`` as a `BaseCorrelatedNoise` object.
        """
        im, scale, var = self.getNoiseProperties(i)
        if im is None:
            cf = UncorrelatedNoise(var, rng=rng, scale=scale, gsparams=gsparams)
        else:
            ii = InterpolatedImage(im, normalization="sb",
                                          calculate_stepk=False, calculate_maxk=False,
                                          x_interpolant='linear', gsparams=gsparams)
            cf = BaseCorrelatedNoise(rng, ii, im.wcs)
            cf = cf.withVariance(var)
        return cf

    def __repr__(self):
        return 'galsim.RealGalaxyCatalog(%r)'%self.file_name

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, RealGalaxyCatalog) and
                 self.file_name == other.file_name and
                 self.image_dir == other.image_dir))
    def __ne__(self, other): return not self.__eq__(other)

    def __hash__(self): return hash(repr(self))

    def __getstate__(self):
        d = self.__dict__.copy()
        d['loaded_files'] = {}
        d['saved_noise_im'] = {}
        del d['gal_lock']
        del d['psf_lock']
        del d['loaded_lock']
        del d['noise_lock']
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        self.gal_lock = Lock()
        self.psf_lock = Lock()
        self.loaded_lock = Lock()
        self.noise_lock = Lock()
        pass

def _parse_files_dirs(file_name, image_dir, sample):
    if sample is None:
        if file_name is None:
            use_sample = '25.2'
        elif '25.2' in file_name:
            use_sample = '25.2'
        elif '23.5' in file_name:
            use_sample = '23.5'
        else:
            use_sample = None
    else:
        use_sample = sample

    if file_name is None:
        file_name = 'real_galaxy_catalog_' + use_sample + '.fits'
        if image_dir is None:
            use_meta_dir = True  # Used to give a more helpful error message
            image_dir = os.path.join(meta_data.share_dir,
                                     'COSMOS_'+use_sample+'_training_sample')
        else:
            use_meta_dir = False
        full_file_name = os.path.join(image_dir,file_name)
        if not os.path.isfile(full_file_name) and use_meta_dir:
            if use_sample not in ('23.5', '25.2'):
                raise GalSimValueError("Sample name not recognized.",use_sample, ('23.5', '25.2'))
            else:
                raise OSError('No RealGalaxy catalog found in %s. Run the program '
                              'galsim_download_cosmos -s %s to download catalog and accompanying '
                              'image files.'%(image_dir, use_sample))
    elif image_dir is None:
        full_file_name = file_name
        image_dir = os.path.dirname(file_name)
    else:
        full_file_name = os.path.join(image_dir,file_name)
    if not os.path.isfile(full_file_name):
        raise OSError(full_file_name+' not found.')

    return full_file_name, image_dir, use_sample


class ChromaticRealGalaxy(ChromaticSum):
    """A class describing real galaxies over multiple wavelengths, using some multi-band training
    dataset.  The underlying implementation models multi-band images of individual galaxies
    as chromatic PSF convolutions (and integrations over wavelength) with a sum of profiles
    separable into spatial and spectral components.  The spectral components are specified by the
    user, and the spatial components are determined one Fourier mode at a time by the class.  This
    decomposition can be thought of as a constrained chromatic deconvolution of the multi-band
    images by the associated PSFs, similar in spirit to `RealGalaxy`.

    Because ChromaticRealGalaxy involves an `InterpolatedKImage`, ``method = 'phot'`` is unavailable
    for the `ChromaticObject.drawImage` function.

    Fundamentally, the required inputs for this class are:

    (1) a series of high resolution input `Image` instances of a single galaxy in different bands,
    (2) a list of `Bandpass` corresponding to those images,
    (3) the PSFs of those images as either `GSObject` or `ChromaticObject` instances, and
    (4) the noise properties of the input images as `BaseCorrelatedNoise` instances.

    If you want to specify these inputs directly, that is possible via the `makeFromImages` factory
    method of this class::

        >>> crg = galsim.ChromaticRealGalaxy.makeFromImages(imgs, bands, PSFs, xis, ...)

    Alternatively, you may create a ChromaticRealGalaxy via a list of `RealGalaxyCatalog` that
    correspond to a set of galaxies observed in different bands::

        >>> crg = galsim.ChromaticRealGalaxy(real_galaxy_catalogs, index=0, ...)

    The above will use the 1st object in the catalogs, which should be the same galaxy, just
    observed in different bands.  Note that there are multiple keywords for choosing a galaxy from
    a catalog; exactly one must be set.  In the future we may add more such options, e.g., to
    choose at random but accounting for the non-constant weight factors (probabilities for
    objects to make it into the training sample).

    The flux normalization of the returned object will by default match the original data, scaled to
    correspond to a 1 second HST exposure (though see the ``area_norm`` parameter).  If you want
    a flux appropriate for a longer exposure or telescope with different collecting area, you can
    use the `ChromaticObject.withScaledFlux` method on the returned object, or use the ``exptime``
    and ``area`` keywords to `ChromaticObject.drawImage`.

    Note that while you can also use `ChromaticObject.withFlux`, `ChromaticObject.withMagnitude`,
    and `ChromaticObject.withFluxDensity` to set the absolute normalization, these methods
    technically adjust the flux of the entire postage stamp image (including noise!) and not
    necessarily the flux of the galaxy itself.  (These two fluxes will be strongly correlated for
    high signal-to-noise ratio galaxies, but may be considerably different at low signal-to-noise
    ratio.)

    Note that ChromaticRealGalaxy objects use arcsec for the units of their linear dimension.  If
    you are using a different unit for other things (the PSF, WCS, etc.), then you should dilate the
    resulting object with ``gal.dilate(galsim.arcsec / scale_unit)``.

    Noise from the original images is propagated by this class, though certain restrictions apply
    to when and how that noise is made available.  The propagated noise depends on which `Bandpass`
    the ChromaticRealGalaxy is being imaged through, so the noise is only available after the
    `ChromaticObject.drawImage` method has been called.  Also, since ChromaticRealGalaxy will
    only produce reasonable images when convolved with a (suitably wide) PSF, the noise attribute is
    attached to the `ChromaticConvolution` (or `ChromaticTransformation` of the
    `ChromaticConvolution`) which holds as one of its convolutants the `ChromaticRealGalaxy`.::

        >>> crg = galsim.ChromaticRealGalaxy(...)
        >>> psf = ...
        >>> obj = galsim.Convolve(crg, psf)
        >>> bandpass = galsim.Bandpass(...)
        >>> assert not hasattr(obj, 'noise')
        >>> image = obj.drawImage(bandpass)
        >>> assert hasattr(obj, 'noise')
        >>> noise1 = obj.noise

    Note that the noise attribute is only associated with the most recently used bandpass.  If you
    draw another image of the same object using a different bandpass, the noise object will be
    replaced.::

        >>> bandpass2 = galsim.Bandpass(...)
        >>> image2 = obj.drawImage(bandpass2)
        >>> assert noise1 != obj.noise

    Parameters:
        real_galaxy_catalogs:   A list of `RealGalaxyCatalog` objects from which to create
                                `ChromaticRealGalaxy` objects.  Each catalog should represent the
                                same set of galaxies, and in the same order, just imaged through
                                different filters.
        index:                  Index of the desired galaxy in the catalog. [One of ``index``,
                                ``id``, or ``random`` is required.]
        id:                     Object ID for the desired galaxy in the catalog. [One of ``index``,
                                ``id``, or ``random`` is required.]
        random:                 If True, then just select a completely random galaxy from the
                                catalog.  [One of ``index``, ``id``, or ``random`` is required.]
        rng:                    A random number generator to use for selecting a random galaxy (may
                                be any kind of `BaseDeviate` or None) and to use in generating any
                                noise field when padding.
        SEDs:                   An optional list of `SED` instances to use when representing real
                                galaxies as sums of separable profiles.  By default, it will use
                                ``len(real_galaxy_catalogs)`` SEDs that are polynomials in
                                wavelength.  Note that if given, ``len(SEDs)`` must equal
                                ``len(real_galaxy_catalogs)``. [default: None]
        k_interpolant:          Either an `Interpolant` instance or a string indicating which
                                k-space interpolant should be used.  Options are 'nearest', 'sinc',
                                'linear', 'cubic', 'quintic', or 'lanczosN' where N should be the
                                integer order to use.  We strongly recommend leaving this parameter
                                at its default value; see text above for details.
                                [default: galsim.Quintic()]
        maxk:                   Optional maxk argument.  If you know you will be convolving the
                                resulting `ChromaticRealGalaxy` with a "fat" PSF in a subsequent
                                step, then it can be more efficient to limit the range of Fourier
                                modes used when solving for the sum of separable profiles below.
                                [default: None]
        pad_factor:             Factor by which to internally oversample the Fourier-space images
                                that represent the `ChromaticRealGalaxy` (equivalent to zero-padding
                                the real-space profiles).  We strongly recommend leaving this
                                parameter at its default value; see text in Realgalaxy docstring
                                for details.  [default: 4]
        noise_pad_size:         If provided, the image will be padded out to this size (in arcsec)
                                with the noise specified in the real galaxy catalog. This is
                                important if you are planning to whiten the resulting image.  You
                                should make sure that the padded image is larger than the postage
                                stamp onto which you are drawing this object.
                                [default: None]
        area_norm:              Area in cm^2 by which to normalize the flux of the returned object.
                                When area_norm=1 (the default), using ``exptime=1`` and ``area=1``
                                arguments in `ChromaticObject.drawImage` (also the default) will
                                simulate an image with the appropriate number of counts for a 1
                                second exposure with the original telescope/camera (e.g., with HST
                                when using the COSMOS catalog).
                                If you would rather explicitly specify the collecting area of the
                                telescope when using `ChromaticObject.drawImage` with a
                                `ChromaticRealGalaxy`, then you should set area_norm equal to the
                                collecting area of the source catalog telescope when creating the
                                `ChromaticRealGalaxy` (e.g., area_norm=45238.93416 for HST).
                                [default: 1]
        gsparams:               An optional `GSParams` argument. [default: None]
        logger:                 A logger object for output of progress statements if the user wants
                                them.  [default: None]

    """
    # TODO: SEDs isn't implemented yet in config parser.
    _opt_params = { "k_interpolant" : str ,
                    "maxk" : float,
                    "pad_factor" : float,
                    "noise_pad_size" : float,
                    "area_norm" : float
                  }
    _single_params = [ { "index" : int , "id" : str , "random" : bool } ]
    _takes_rng = True

    def __init__(self, real_galaxy_catalogs, index=None, id=None, random=False, rng=None,
                 gsparams=None, logger=None, **kwargs):
        if rng is None:
            rng = BaseDeviate()
        elif not isinstance(rng, BaseDeviate):
            raise TypeError("The rng provided to ChromaticRealGalaxy is not a BaseDeviate")
        self.rng = rng

        logger = LoggerWrapper(logger)  # So don't need to check `if logger:` all the time.

        # Get the index to use in the catalog
        if index is not None:
            if id is not None or random:
                raise GalSimIncompatibleValuesError(
                    "Too many methods for selecting a galaxy.", index=index, id=id, random=random)
            use_index = index
        elif id is not None:
            if random:
                raise GalSimIncompatibleValuesError(
                    "Too many methods for selecting a galaxy.", id=id, random=random)
            use_index = real_galaxy_catalogs[0].getIndexForID(id)
        elif random:
            uniform_deviate = UniformDeviate(self.rng)
            use_index = int(real_galaxy_catalogs[0].nobjects * uniform_deviate())
        else:
            raise GalSimIncompatibleValuesError(
                "No method specified for selecting a galaxy.", index=index, id=id, random=random)
        logger.debug('ChromaticRealGalaxy %d: Start ChromaticRealGalaxy constructor.', use_index)
        self.index = use_index

        # Read in the galaxy, PSF images; for now, rely on pyfits to make I/O errors.
        imgs = [rgc.getGalImage(use_index) for rgc in real_galaxy_catalogs]
        logger.debug('ChromaticRealGalaxy %d: Got gal_image', use_index)

        PSFs = [rgc.getPSF(use_index) for rgc in real_galaxy_catalogs]
        logger.debug('ChromaticRealGalaxy %d: Got psf', use_index)

        bands = [rgc.getBandpass() for rgc in real_galaxy_catalogs]

        xis = []
        for rgc in real_galaxy_catalogs:
            noise_image, pixel_scale, var = rgc.getNoiseProperties(use_index)
            # Make sure xi image is odd-sized.
            if noise_image.array.shape[0] % 2 == 0: #pragma: no branch
                bds = noise_image.bounds
                new_bds = BoundsI(bds.xmin+1, bds.xmax, bds.ymin+1, bds.ymax)
                noise_image = noise_image[new_bds]
            ii = InterpolatedImage(noise_image, normalization='sb',
                                   calculate_stepk=False, calculate_maxk=False,
                                   x_interpolant='linear', gsparams=gsparams)
            xi = BaseCorrelatedNoise(self.rng, ii, noise_image.wcs)
            xi = xi.withVariance(var)
            xis.append(xi)
        logger.debug('ChromaticRealGalaxy %d: Got noise_image',use_index)
        self.catalog_files = [rgc.getFileName() for rgc in real_galaxy_catalogs]

        self._initialize(imgs, bands, xis, PSFs, gsparams=gsparams, **kwargs)

    @classmethod
    def makeFromImages(cls, images, bands, PSFs, xis, **kwargs):
        """Create a `ChromaticRealGalaxy` directly from images, bandpasses, PSFs, and noise
        descriptions.  See the `ChromaticRealGalaxy` docstring for more information.

        Parameters:
            images:             An iterable of high resolution `Image` instances of a galaxy
                                through different bandpasses.
            bands:              An iterable of `Bandpass` objects corresponding to the  input
                                images.
            PSFs:               Either an iterable of `GSObject` or `ChromaticObject` indicating
                                the PSFs of the different input images, or potentially a single
                                `GSObject` or `ChromaticObject` that will be used as the PSF for
                                all images.
            xis:                An iterable of `BaseCorrelatedNoise` objects characterizing the
                                noise in the input images.
            SEDs:               An optional list of `SED` instances to use when representing real
                                galaxies as sums of separable profiles.  By default, it will use
                                ``len(images)`` SEDs that are polynomials in wavelength.  Note that
                                if given, ``len(SEDs)`` must equal ``len(images)``. [default: None]
            k_interpolant:      Either an `Interpolant` instance or a string indicating which
                                k-space interpolant should be used.  Options are 'nearest', 'sinc',
                                'linear', 'cubic', 'quintic', or 'lanczosN' where N should be the
                                integer order to use.  We strongly recommend leaving this parameter
                                at its default value; see text above for details.  [default:
                                galsim.Quintic()]
            maxk:               Optional maxk argument.  If you know you will be convolving the
                                resulting `ChromaticRealGalaxy` with a "fat" PSF in a subsequent
                                step, then it can be more efficient to limit the range of Fourier
                                modes used when solving for the sum of separable profiles below.
                                [default: None]
            pad_factor:         Factor by which to internally oversample the Fourier-space images
                                that represent the `ChromaticRealGalaxy` (equivalent to zero-padding
                                the real-space profiles).  We strongly recommend leaving this
                                parameter at its default value; see text in Realgalaxy docstring
                                for details.  [default: 4]
            noise_pad_size:     If provided, the image will be padded out to this size (in arcsec)
                                with the noise specified in the real galaxy catalog. This is
                                important if you are planning to whiten the resulting image.  You
                                should make sure that the padded image is larger than the postage
                                stamp onto which you are drawing this object.
                                [default: None]
            area_norm:          Area in cm^2 by which to normalize the flux of the returned object.
                                When area_norm=1 (the default), using ``exptime=1`` and ``area=1``
                                arguments in `ChromaticObject.drawImage` (also the default) will
                                simulate an image with the appropriate number of counts for a 1
                                second exposure with the original telescope/camera (e.g., with HST
                                when using the COSMOS catalog).
                                If you would rather explicitly specify the collecting area of the
                                telescope when using `ChromaticObject.drawImage` with a
                                `ChromaticRealGalaxy`, then you should set area_norm equal to the
                                collecting area of the source catalog telescope when creating the
                                `ChromaticRealGalaxy` (e.g., area_norm=45238.93416 for HST).
                                [default: 1]
            gsparams:           An optional `GSParams` argument. [default: None]
            logger:             A logger object for output of progress statements if the user wants
                                them.  [default: None]

        """
        if not hasattr(PSFs, '__iter__'):
            PSFs = [PSFs]*len(images)
        obj = cls.__new__(cls)
        obj.index = None
        obj.catalog_files = None
        obj.rng = kwargs.pop('rng', BaseDeviate())

        if len(images) != len(bands) or len(images) != len(xis) or len(images) != len(PSFs):
            raise GalSimIncompatibleValuesError(
                "The number of images, bands, xis, and PSFs must match.",
                images=images, bands=bands, xis=xis, PSFs=PSFs)

        obj._initialize(images, bands, xis, PSFs, **kwargs)
        return obj

    def _initialize(self, imgs, bands, xis, PSFs,
                    SEDs=None, k_interpolant=None, maxk=None, pad_factor=4., area_norm=1.0,
                    noise_pad_size=0, gsparams=None):

        if SEDs is None:
            SEDs = self._poly_SEDs(bands)
        elif len(SEDs) > len(imgs):
            raise GalSimIncompatibleValuesError(
                "The number of SEDs must be <= the number of images",
                images=imgs, SEDs=SEDs)
        self.SEDs = SEDs

        if k_interpolant is None:
            k_interpolant = Quintic()
        else:
            k_interpolant = convert_interpolant(k_interpolant)

        self._area_norm = area_norm
        self._k_interpolant = k_interpolant
        self._gsparams = GSParams.check(gsparams)

        NSED = len(self.SEDs)
        Nim = len(imgs)
        #assert Nim == len(bands)
        #assert Nim == len(xis)
        #assert Nim == len(PSFs)
        #assert Nim >= NSED

        if area_norm != 1.0:
            imgs = [img/area_norm for img in imgs]
            xis = [xi/area_norm**2 for xi in xis]

        # Need to sample three different types of objects on the same Fourier grid: the input
        # effective PSFs, the input images, and the input correlation-functions/power-spectra.
        # There are quite a few potential options for implementing this Fourier sampling.  Some
        # examples include:
        #   * draw object in real space, interpolate onto the real-space grid conjugate to the
        #     desired Fourier-space grid and then DFT with numpy.fft methods.
        #   * Use numpy.fft methods on pre-sampled real-space input (like the input images), then
        #     use an InterpolatedKImage object to regrid onto desired Fourier grid.
        #   * Create an InterpolatedImage from pre-sampled input then use drawKImage to directly
        #     sample on desired Fourier grid.
        # I'm sure there are other options too.  The options chosen below were chosen empirically
        # based on tests of propagating both (chromatic) galaxy images and images of pure noise.

        # Select maxk by requiring modes to be resolved both by the marginal PSFs (i.e., the
        # achromatic PSFs obtained by evaluating the chromatic PSF at the blue and red edges of
        # each of the filters provided) and also by the input images' pixel scales.

        img_maxk = np.min([np.pi/img.scale for img in imgs])
        marginal_PSFs = [PSF.evaluateAtWavelength(band.blue_limit)
                         for PSF in PSFs for band in bands]
        marginal_PSFs += [PSF.evaluateAtWavelength(band.red_limit)
                          for PSF in PSFs for band in bands]
        psf_maxk = np.min([p.maxk for p in marginal_PSFs])

        # In practice, the output PSF should almost always cut off at smaller maxk than obtained
        # above.  In this case, the user can set the maxk keyword argument for improved efficiency.
        if maxk is None:
            maxk = np.min([img_maxk, psf_maxk])
        else:
            maxk = np.min([img_maxk, psf_maxk, maxk])

        # Setting stepk is trickier.  We'll assume that the postage stamp inputs are already at the
        # critical size to avoid significant aliasing and use the implied stepk.  We'll insist that
        # the WCS is a simple PixelScale.  We'll also use the same trick that InterpolatedImage
        # uses to improve accuracy, namely, increase the Fourier-space resolution a factor of
        # `pad_factor`.
        stepk = np.min([2*np.pi/(img.scale*max(img.array.shape))/pad_factor for img in imgs])
        nk = 2*int(np.floor(maxk/stepk))

        # Create Fourier-space kimages of effective PSFs
        PSF_eff_kimgs = np.empty((Nim, NSED, nk, nk), dtype=np.complex128)
        for i, (img, band, PSF) in enumerate(zip(imgs, bands, PSFs)):
            for j, sed in enumerate(self.SEDs):
                # assume that PSF already includes pixel, so don't convolve one in again.
                PSF_eff_kimgs[i, j] = (PSF * sed).drawKImage(band, nx=nk, ny=nk, scale=stepk).array

        # Get Fourier-space representations of input imgs.
        kimgs = np.empty((Nim, nk, nk), dtype=np.complex128)

        if noise_pad_size == 0:
            noise_pad = 0.

        for i, (img, xi) in enumerate(zip(imgs, xis)):
            if noise_pad_size != 0:
                noise_pad = xi
            ii = InterpolatedImage(img, noise_pad_size=noise_pad_size, noise_pad=noise_pad,
                                   rng=self.rng, pad_factor=pad_factor)
            kimgs[i] = ii.drawKImage(nx=nk, ny=nk, scale=stepk).array

        # Setup input noise power spectra
        pks = np.empty((Nim, nk, nk), dtype=np.float64)
        for i, (img, xi) in enumerate(zip(imgs, xis)):
            pks[i] = xi.drawKImage(nx=nk, ny=nk, scale=stepk).array.real / xi.wcs.pixelArea()
            ny, nx = img.array.shape
            pks[i] *= nx * ny
        w = 1./np.sqrt(pks)

        # Allocate and fill output coefficients and covariances.
        # Note: put NSED axis last, since significantly faster to compute them this way,
        # even though we eventually convert to images which are strided in this format.
        coef = np.zeros((nk, nk, NSED), dtype=np.complex128)
        Sigma = np.empty((nk, nk, NSED, NSED), dtype=np.complex128)

        # Solve the weighted linear least squares problem for each Fourier mode.  This is
        # effectively a constrained chromatic deconvolution.  Take advantage of symmetries.
        _coef = coef.__array_interface__['data'][0]
        _Sigma = Sigma.__array_interface__['data'][0]
        _w = w.__array_interface__['data'][0]
        _kimgs = kimgs.__array_interface__['data'][0]
        _psf = PSF_eff_kimgs.__array_interface__['data'][0]
        _galsim.ComputeCRGCoefficients(_coef, _Sigma, _w, _kimgs, _psf, NSED, Nim, nk, nk)

        # Reorder these so they correspond to (NSED, nky, nkx) and (NSED, NSED, nky, nkx) shapes.
        coef = np.transpose(coef, (2,0,1))
        Sigma = np.transpose(Sigma, (2,3,0,1))

        # Set up obj_list as required of ChromaticSum subclass.
        obj_list = []
        for i, sed in enumerate(self.SEDs):
            obj_list.append(sed * _InterpolatedKImage(
                    ImageCD(coef[i], scale=stepk),
                    k_interpolant=self._k_interpolant,
                    gsparams=self._gsparams))

        Sigma_dict = {}
        for i in range(NSED):
            for j in range(i, NSED):
                obj = _InterpolatedKImage(
                        ImageCD(Sigma[i, j], scale=stepk),
                        k_interpolant=self._k_interpolant,
                        gsparams=self._gsparams)
                obj /= (imgs[0].array.shape[0] * imgs[0].array.shape[1] * imgs[0].scale**2)
                Sigma_dict[(i, j)] = obj

        self.covspec = CovarianceSpectrum(Sigma_dict, self.SEDs)

        ChromaticSum.__init__(self, obj_list)

    @staticmethod
    def _poly_SEDs(bands):
        # Use polynomial SEDs by default; up to the number of bands provided.
        waves = []
        for bp in bands:
            waves = merge_sorted([waves, bp.wave_list])
        SEDs = []
        for i in range(len(bands)):
            SEDs.append(
                    SED(LookupTable(waves, waves**i, interpolant='linear'), 'nm', 'fphotons')
                    .withFlux(1.0, bands[0]))
        return SEDs

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, ChromaticRealGalaxy) and
                 self.catalog_files == other.catalog_files and
                 self.index == other.index and
                 self.SEDs == other.SEDs and
                 self._k_interpolant == other._k_interpolant and
                 self._area_norm == other._area_norm and
                 self._gsparams == other._gsparams))
    def __ne__(self, other): return not self.__eq__(other)

    def __hash__(self):
        return hash(("galsim.ChromaticRealGalaxy", tuple(self.catalog_files), self.index,
                     tuple(self.SEDs), self._k_interpolant, self._area_norm, self._gsparams))

    def __str__(self):
        return "galsim.ChromaticRealGalaxy(%r, index=%r)"%(self.catalog_files, self.index)

    def __repr__(self):
        return ("galsim.ChromaticRealGalaxy(%r, SEDs=%r, index=%r, k_interpolant=%r, "
                "area_norm=%r, gsparams=%r)"%(self.catalog_files, self.SEDs, self.index,
                                              self._k_interpolant, self._area_norm, self._gsparams))

# Put this at the bottom to avoid circular import error.
from .config import LoggerWrapper




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from .extra import ExtraOutputBuilder, RegisterExtraOutput
from .value import ParseValue
from .noise import AddNoiseVariance
from ..image import ImageF

# The weight extra output type builds an ImageF of the inverse noise variance in the image.
# It builds up the variance either from the stamp information if noise is being added then
# or at the end from the full image if that is when noise is added.  Then at the end of
# the image processing, it inverts the image to get the appropriate weight map.

class WeightBuilder(ExtraOutputBuilder):
    """This builds a weight map image to go along with each regular data image.

    The weight is the inverse variance of the noise in the image.

    There is a option called 'include_obj_var' that governs whether the weight should include the
    Poisson variance of the signal.  In real data, you don't know the true signal, and estimating
    the Poisson noise from the realized image can lead to biases.  As such, different applications
    may or may not want this included.
    """

    # The function to call at the end of building each stamp
    def processStamp(self, obj_num, config, base, logger):
        if base['do_noise_in_stamps']:
            weight_im = ImageF(base['current_stamp'].bounds, wcs=base['wcs'], init_value=0.)
            if 'include_obj_var' in config:
                include_obj_var = ParseValue(config, 'include_obj_var', base, bool)[0]
            else:
                include_obj_var = False
            base['current_noise_image'] = base['current_stamp']
            AddNoiseVariance(base,weight_im,include_obj_var,logger)
            self.scratch[obj_num] = weight_im

    # The function to call at the end of building each image
    def processImage(self, index, obj_nums, config, base, logger):
        image = ImageF(base['image_bounds'], wcs=base['wcs'], init_value=0.)
        if len(self.scratch) > 0.:
            # If we have been accumulating the variance on the stamps, build the total from them.
            for obj_num in obj_nums:
                stamp = self.scratch[obj_num]
                b = stamp.bounds & image.bounds
                if b.isDefined(): # pragma: no branch
                    # This next line is equivalent to:
                    #    image[b] += stamp[b]
                    # except that this doesn't work through the proxy.  We can only call methods
                    # that don't start with _.  Hence using the more verbose form here.
                    image.setSubImage(b, image.subImage(b) + stamp[b])
        else:
            # Otherwise, build the variance map now.
            if 'include_obj_var' in config:
                include_obj_var = ParseValue(config, 'include_obj_var', base, bool)[0]
            else:
                include_obj_var = False
            base['current_noise_image'] = base['current_image']
            AddNoiseVariance(base,image,include_obj_var,logger)

        # Now invert the variance image to get weight map.
        # Note that any zeros present in the image are maintained as zeros after inversion.
        # So it is ok to set bad pixels to have zero variance above, and they will invert to have
        # zero weight.
        image.invertSelf()
        self.data[index] = image


# Register this as a valid extra output
RegisterExtraOutput('weight', WeightBuilder())
